#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @brief Base class for configuration
# @authors Adrien Crovato

from __future__ import print_function
from builtins import range
from builtins import object
import flow as f
import tbox
import tbox.gmsh as gmsh
import fwk
from fwk.wutils import parseargs
from fwk.coloring import ccolors

class Config(object):
    def __init__(self, p):
        # timer
        self.tms = fwk.Timers()
        self.tms["total"].start()

        # basic checks
        if p['Dim'] != 2 and p['Dim'] != 3:
            raise Exception('Problem dimension should be 2 or 3, but ' + p['Dim'] + ' was given!\n')
        # basic config
        if p['Format'] == 'vtk':
            try:   
               import tboxVtk
               Writer = tboxVtk.VtkExport
               print("Found VTK libraries! Results will be saved in VTK format.\n")
            except:
               Writer = tbox.GmshExport
               print("VTK libraries not found! Results will be saved in gmsh format.\n")
        else:
            Writer = tbox.GmshExport
            print("Results will be saved in gmsh format.\n")

        # mesh the airfoil
        print(ccolors.ANSI_BLUE + 'Meshing the airfoil...' + ccolors.ANSI_RESET)
        self.msh = gmsh.MeshLoader(p['File'],__file__).execute(**p['Pars'])
        if p['Dim'] == 2:
            mshCrck = tbox.MshCrack(self.msh, p['Dim'])
            mshCrck.setCrack(p['Wake'])
            mshCrck.addBoundaries([p['Fluid'], p['Farfield'][-1], p['Wing']])
            mshCrck.run()
        else:
            for i in range(0, len(p['Wakes'])):
                mshCrck = tbox.MshCrack(self.msh, p['Dim'])
                mshCrck.setCrack(p['Wakes'][i])
                if 'Fuselage' in p:
                    mshCrck.addBoundaries([p['Fluid'], p['Symmetry'], p['Farfield'][-1], p['Fuselage'], p['Wings'][i]])
                else:
                    mshCrck.addBoundaries([p['Fluid'], p['Symmetry'], p['Farfield'][-1], p['Wings'][i]])
                mshCrck.setExcluded(p['WakeTips'][i])
                mshCrck.run()
        tbox.GmshExport(self.msh).save(self.msh.name)
        del mshCrck
        self.mshWriter = Writer(self.msh)
        print('\n')

        # initialize the problem
        self.pbl = f.Problem(self.msh, p['Dim'], 0., p['M_inf'], p['S_ref'], p['c_ref'], p['x_ref'], p['y_ref'], p['z_ref'])
        self.phiInfFun = f.F0PsPhiInf(p['Dim'], 0.)
        self.velInfFun = tbox.Fct1C(0., 0., 0.)
        # add a medium "air"
        if p['M_inf'] == 0:
            self.pbl.set(f.Medium(self.msh, p['Fluid'], f.F0ElRhoL(), f.F0ElMachL(), f.F0ElCpL(), self.phiInfFun))
        else:
            self.pbl.set(f.Medium(self.msh, p['Fluid'], f.F0ElRho(p['M_inf'], p['M_crit']), f.F0ElMach(p['M_inf']), f.F0ElCp(p['M_inf']), self.phiInfFun))
        # add initial condition
        self.pbl.add(f.Initial(self.msh, p['Fluid'], self.phiInfFun))
        # add farfield boundary conditions
        self.pbl.add(f.Dirichlet(self.msh, p['Farfield'][0], self.phiInfFun))
        for i in range (1, len(p['Farfield'])):
            self.pbl.add(f.Freestream(self.msh, p['Farfield'][i], self.velInfFun))
        # add solid boundaries
        if p['Dim'] == 2:
            self.bnd = f.Boundary(self.msh, [p['Wing'], p['Fluid']])
            self.pbl.add(self.bnd)
        else:
            self.bnd = []
            for bd in p['Wings']:
                bnd = f.Boundary(self.msh, [bd, p['Fluid']])
                self.bnd.append(bnd)
                self.pbl.add(bnd)
            if 'Fuselage' in p:
                bnd = f.Boundary(self.msh, [p['Fuselage'], p['Fluid']])
                self.bnd.append(bnd)
                self.pbl.add(bnd)
        # add Wake/Kutta boundary conditions
        if p['Dim'] == 2:
            self.pbl.add(f.Wake(self.msh, [p['Wake'], p['Wake']+'_', p['Fluid']]))
            self.pbl.add(f.Kutta(self.msh, [p['Te'], p['Wake']+'_', p['Wing'], p['Fluid']]))
        else:
            for i in range(0, len(p['Wakes'])):
                self.pbl.add(f.Wake(self.msh, [p['Wakes'][i], p['Wakes'][i]+'_', p['Fluid'], p['TeTips'][i]]))

        # initialize the solver
        print(ccolors.ANSI_BLUE + 'Initializing the solver...' + ccolors.ANSI_RESET)
        if p['NSolver'] == 'Picard':
            self.solver = f.Picard(self.pbl)
            self.solver.relax = p['Relaxation']
        elif p['NSolver'] == 'Newton':
            self.solver = f.Newton(self.pbl)
            self.solver.lsTol = p['LS_tol']
            self.solver.maxLsIt = p['Max_it_LS']
            self.solver.avThrsh = p['AV_thrsh']
        else:
            raise RuntimeError('Available nonlinear solver type: Picard or Newton, but ' + p['NSolver'] + ' was given!\n')
        args = parseargs()
        self.solver.nthreads = args.k
        self.solver.relTol = p['Rel_tol']
        self.solver.absTol = p['Abs_tol']
        self.solver.maxIt = p['Max_it']
