/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wAdjoint.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wAssign.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wBoundary.h"
#include "wSolver.h"

#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTag.h"
#include "wPt.h"
#include "wCache.h"
#include "wMem.h"
#include "wResults.h"
#include "wMshExport.h"

#include "gmm_gmm.h"
#include <gmm/gmm_MUMPS_interface.h>
#include "gmm_extra.h"
#include "std_extra.h"

#include <tbb/task_scheduler_init.h>
#include <tbb/tbb.h>
#include <tbb/parallel_do.h>
#include <tbb/spin_mutex.h>

#include <iomanip>

using namespace tbox;
using namespace flow;

#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

/**
 * @brief Initialize the adjoint solver
 * @authors Adrien Crovato
 */
Adjoint::Adjoint(std::shared_ptr<Solver> _sol) : sol(_sol)
{
    // Default parameters
    nthreads = sol->nthreads;
    verbose = sol->verbose;

    // Setup variables (unkwown and residual vectors)
    lambdaL.resize(sol->pbl->msh->nodes.size(), 0.);
    lambdaD.resize(sol->pbl->msh->nodes.size(), 0.);
}

/**
 * @brief Run the linear solver
 * 
 * Solve the adjoint steady transonic Full Potential Equation
 * i.e. computes sensitivites for lift and drag cost functions
 * @authors Adrien Crovato
 */
void Adjoint::run()
{
    // Init
    tbb::task_scheduler_init init(nthreads);

    // Display solver parameters
    std::cout << "--- Adjoint solver ---\n"
              << "Number of threads: " << nthreads << "\n"
              << std::endl;
    // build SoE
    gmm::csr_matrix<double> dR;
    buildDR(dR);
    std::vector<double> dL(sol->pbl->msh->nodes.size()), dD(sol->pbl->msh->nodes.size());
    buildDI(dL, dD);

    // Solve linear SoE dR {lambdaL, lambdaD} = {dL, dD}
    bool solve_success = gmm::MUMPS_solve(dR, lambdaL, dL);
    if (!solve_success)
        throw std::runtime_error("Problem occured in gmm::MUMPS_solve while solving for lift sensitivities!");
    solve_success = gmm::MUMPS_solve(dR, lambdaD, dD);
    if (!solve_success)
        throw std::runtime_error("Problem occured in gmm::MUMPS_solve while solving for drag sensitivities!");

    // Check residual
    std::vector<double> rLambdaL(sol->pbl->msh->nodes.size()), rLambdaD(sol->pbl->msh->nodes.size()), dR_lambdaL(sol->pbl->msh->nodes.size()), dR_lambdaD(sol->pbl->msh->nodes.size());
    gmm::mult(dR, lambdaL, dR_lambdaL);
    gmm::mult(dR, lambdaD, dR_lambdaD);
    gmm::add(dR_lambdaL, gmm::scaled(dL, -1.0), rLambdaL);
    gmm::add(dR_lambdaD, gmm::scaled(dD, -1.0), rLambdaD);

    std::cout << std::setw(8) << "L_Iter"
              << std::setw(15) << "Sens[lambdaL]"
              << std::setw(15) << "Sens[lambdaD]"
              << std::setw(15) << "Res[lambdaL]"
              << std::setw(15) << "Res[lambdaD]" << std::endl;
    std::cout << std::fixed << std::setprecision(5);
    std::cout << std::setw(8) << 0
              << std::setw(15) << gmm::vect_norm2(lambdaL)
              << std::setw(15) << gmm::vect_norm2(lambdaD)
              << std::setw(15) << log10(gmm::vect_norm2(rLambdaL))
              << std::setw(15) << log10(gmm::vect_norm2(rLambdaD)) << std::endl;

    std::cout << ANSI_COLOR_YELLOW << "Warning: the adjoint solver is experimental and has not been validated!\n"
              << ANSI_COLOR_RESET << std::endl;
}

/**
 * @brief Build the Jacobian matrix
 * @authors Adrien Crovato
 * @todo Refactor, this is heavily copy-pasted from Newton::buildJ
 */
void Adjoint::buildDR(gmm::csr_matrix<double> &dR)
{
    // Multithread
    tbb::spin_mutex mutex;

    // Temporary Jacobian matrix
    gmm::row_matrix<gmm::wsvector<double>> A(sol->pbl->msh->nodes.size(), sol->pbl->msh->nodes.size());

    // Full Potential Equation with upwind bias: analytical derivatives
    auto fluid = sol->pbl->medium;
    tbb::parallel_do(fluid->adjMap.begin(), fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> p) {
        // Current element
        Element *e = p.first;
        //Upwind element
        Element *eU = p.second[0];

        // Subsonic contribution (drho*grad_phi*grad_psi + rho*grad_phi*grad_psi)
        gmm::dense_matrix<double> Ae1(e->nodes.size(), e->nodes.size());
        e->buildDK(Ae1, sol->phi, fluid->rho);
        gmm::dense_matrix<double> Ae2(e->nodes.size(), e->nodes.size());
        e->buildK(Ae2, sol->phi, fluid->rho);
        // Supersonic contribution
        double mCOv = 0.95; // solver should be converged, so we set the lowest numerical viscosity
        double muCv = 1.0;
        double mach = fluid->mach.eval(*e, sol->phi, 0);
        if (mach > mCOv)
        {
            // switching function and derivative
            double mu = muCv * (1 - (mCOv * mCOv) / (mach * mach));
            double dmu = 2 * muCv * mCOv * mCOv / (mach * mach * mach);

            // scale 1st and 2nd terms
            gmm::scale(Ae1, 1 - mu);
            gmm::scale(Ae2, 1 - mu);

            // 3rd term (mu*drhoU*grad_phi*grad_psi)
            std::vector<double> bei(e->nodes.size());
            std::vector<double> bejU(eU->nodes.size());
            e->buildDs(bei, sol->phi, Fct0C(1.));
            std::vector<double> gradU(sol->pbl->nDim);
            eU->computeGrad(sol->phi, gradU, 0);
            std::vector<double> dNU(sol->pbl->nDim);
            Cache &cacheU = eU->getVCache();
            gmm::dense_matrix<double> JinvU = eU->getVMem().getJinv(0);
            for (size_t j = 0; j < eU->nodes.size(); ++j)
            {
                gmm::mult(JinvU, gmm::col_vector(cacheU.dff[0][j]), gmm::col_vector(dNU));
                for (size_t i = 0; i < sol->pbl->nDim; ++i)
                    bejU[j] += gradU[i] * dNU[i];
            }
            gmm::dense_matrix<double> Ae3(e->nodes.size(), eU->nodes.size());
            for (size_t i = 0; i < e->nodes.size(); ++i)
                for (size_t j = 0; j < eU->nodes.size(); ++j)
                    Ae3(i, j) = mu * fluid->rho.evalD(*eU, sol->phi, 0) * bei[i] * bejU[j];

            // 4th term (mu*rhoU*grad_phi*grad_psi)
            gmm::dense_matrix<double> Ae4(e->nodes.size(), e->nodes.size());
            e->buildK(Ae4, sol->phi, Fct0C(1.));
            gmm::scale(Ae4, mu * fluid->rho.eval(*eU, sol->phi, 0));

            // 5th term dmu*(rhoU-rho)*grad_phi*grad_psi
            double deltaRho = fluid->rho.eval(*eU, sol->phi, 0) - fluid->rho.eval(*e, sol->phi, 0);
            gmm::dense_matrix<double> Ae5(e->nodes.size(), e->nodes.size());
            e->buildDK(Ae5, sol->phi, fluid->mach);
            gmm::scale(Ae5, deltaRho * dmu);

            // Assembly (supersonic)
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < e->nodes.size(); ++ii)
            {
                Node *nodi = e->nodes[ii];
                for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                {
                    Node *nodj = e->nodes[jj];
                    A(nodi->row, nodj->row) += Ae4(ii, jj) + Ae5(ii, jj);
                }
                for (size_t jj = 0; jj < eU->nodes.size(); ++jj)
                {
                    Node *nodj = eU->nodes[jj];
                    A(nodi->row, nodj->row) += Ae3(ii, jj);
                }
            }
        }
        // Assembly (subsonic)
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t ii = 0; ii < e->nodes.size(); ++ii)
        {
            Node *nodi = e->nodes[ii];
            for (size_t jj = 0; jj < e->nodes.size(); ++jj)
            {
                Node *nodj = e->nodes[jj];
                A(nodi->row, nodj->row) += Ae1(ii, jj) + Ae2(ii, jj);
            }
        }
    });
    // Apply wake BCs for adjoint problem
    for (auto wake : sol->pbl->wBCs)
    {
        tbb::parallel_do(wake->wEle.begin(), wake->wEle.end(), [&](WakeElement *we) {
            gmm::dense_matrix<double> Kupup(we->nColUp, we->nRow);
            gmm::dense_matrix<double> Kuplw(we->nColUp, we->nRow);
            gmm::dense_matrix<double> Klwup(we->nColLw, we->nRow);
            gmm::dense_matrix<double> Klwlw(we->nColLw, we->nRow);
            buildWake(we, Kupup, Kuplw, Klwup, Klwlw);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < we->nColUp; ++ii)
            {
                Node *nodi = we->volUpE->nodes[ii];
                for (size_t jj = 0; jj < we->nRow; ++jj)
                {
                    Node *nodj = we->surUpE->nodes[jj];
                    A(nodi->row, nodj->row) += 2 * Kupup(ii, jj);
                    nodj = we->surLwE->nodes[jj];
                    A(nodi->row, nodj->row) += Kuplw(ii, jj);
                }
            }
            for (size_t ii = 0; ii < we->nColLw; ++ii)
            {
                Node *nodi = we->volLwE->nodes[ii];
                for (size_t jj = 0; jj < we->nRow; ++jj)
                {
                    Node *nodj = we->surUpE->nodes[jj];
                    A(nodi->row, nodj->row) -= 2 * Klwup(ii, jj);
                    nodj = we->surLwE->nodes[jj];
                    A(nodi->row, nodj->row) -= Klwlw(ii, jj);
                }
            }
        });
    }
    // Apply Dirichlet BCs
    for (auto dBC : sol->pbl->dBCs)
    {
        for (auto nod : dBC->nodes)
        {
            A.row(nod->row) = gmm::wsvector<double>(sol->pbl->msh->nodes.size());
            A(nod->row, nod->row) = 1.0;
        }
    }
    for (auto fBC : sol->pbl->fBCs)
    {
        for (auto e : fBC->tag->elems)
        {
            for (auto nod : e->nodes)
            {
                A.row(nod->row) = gmm::wsvector<double>(sol->pbl->msh->nodes.size());
                A(nod->row, nod->row) = 1.0;
            }
        }
    }
    // Transpose
    gmm::transposed(A);
    gmm::copy(A, dR);

    if (verbose)
        std::cout << "dR (" << gmm::mat_nrows(dR) << "," << gmm::mat_ncols(dR) << ") nnz=" << gmm::nnz(dR) << "\n";
}

/**
 * @brief Build the derivatives of the cost functions
 * @authors Adrien Crovato
 * @todo Refactor, to be checked
 */
void Adjoint::buildDI(std::vector<double> &dL, std::vector<double> &dD)
{
    // Multithread
    tbb::spin_mutex mutex;

    // Derivative of lift and drag
    for (auto sur : sol->pbl->bnd)
    {
        tbb::parallel_do(sur->groups[0]->tag->elems.begin(), sur->groups[0]->tag->elems.end(), [&](Element *e) {
            // Build flux factor
            Element *eV = sur->svMap.at(e);
            std::vector<double> be(eV->nodes.size(), 0.);
            buildBoundary(e, eV, be);
            // Compute scaling factor
            Pt Vi(0., 0., 0.);
            Vi.x[0] = -sin(sol->pbl->alpha);
            Vi.x[sol->pbl->nDim - 1] = cos(sol->pbl->alpha);
            double factorL = -1 / sol->pbl->S_ref * sol->pbl->medium->cP.evalD(*eV, sol->phi, 0) * e->normal() * Vi;
            Vi.x[0] = cos(sol->pbl->alpha);
            Vi.x[sol->pbl->nDim - 1] = sin(sol->pbl->alpha);
            double factorD = -1 / sol->pbl->S_ref * sol->pbl->medium->cP.evalD(*eV, sol->phi, 0) * e->normal() * Vi;

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < eV->nodes.size(); ++ii)
            {
                Node *nodi = eV->nodes[ii];
                dL[nodi->row] += factorL * be[ii];
                dD[nodi->row] += factorD * be[ii];
            }
        });
    }
    // Apply Dirichlet BCs, adjoint solution vanishes at infinity
    for (auto dBC : sol->pbl->dBCs)
    {
        for (auto nod : dBC->nodes)
        {
            dL[nod->row] = 0.;
            dD[nod->row] = 0.;
        }
    }
    for (auto fBC : sol->pbl->fBCs)
    {
        for (auto e : fBC->tag->elems)
        {
            for (auto nod : e->nodes)
            {
                dL[nod->row] = 0.;
                dD[nod->row] = 0.;
            }
        }
    }

    if (verbose)
    {
        std::cout << "dL (" << gmm::vect_size(dL) << ")\n";
        std::cout << "dD (" << gmm::vect_size(dD) << ")\n";
    }
}

/**
 * @brief Build the boundary contribution for the adjoint equation
 * @authors Adrien Crovato
 * @todo Refactor, to be checked and moved to Boundary::buildAdjNs()
 */
void Adjoint::buildBoundary(Element *&e, Element *&eV, std::vector<double> &be)
{
    // Get shape functions
    Cache &volCache = eV->getVCache();
    // Get Jacobian
    Mem &surMem = e->getVMem();
    Mem &volMem = eV->getVMem();
    // Velocity
    std::vector<double> grad(sol->pbl->nDim);
    eV->computeGrad(sol->phi, grad, 0);
    gmm::dense_matrix<double> V(1, sol->pbl->nDim);
    for (size_t i = 0; i < sol->pbl->nDim; ++i)
        V(0, i) = grad[i];
    // dN
    size_t nCol = eV->nodes.size();
    gmm::dense_matrix<double> dN(sol->pbl->nDim, nCol);
    for (size_t j = 0; j < nCol; ++j)
    {
        for (size_t i = 0; i < sol->pbl->nDim; ++i)
        {
            dN(i, j) = volCache.dff[0][j][i];
        }
    }
    // inv(J)*dN
    gmm::dense_matrix<double> JdN(sol->pbl->nDim, nCol);
    gmm::mult(volMem.getJinv(0), dN, JdN);
    // b = V*[inv(J)*dN]
    gmm::dense_matrix<double> B(1, nCol);
    gmm::mult(V, JdN, B);
    for (size_t i = 0; i < nCol; ++i)
        be[i] = B(0, i);
    gmm::scale(be, surMem.getVol());
}

/**
 * @brief Build the wake contribution for the adjoint equation
 * @authors Adrien Crovato
 * @todo Refactor, to be checked and moved to WakeElement::buildAdjNK()
 */
void Adjoint::buildWake(WakeElement *&we, gmm::dense_matrix<double> &Kupup, gmm::dense_matrix<double> &Kuplw, gmm::dense_matrix<double> &Klwup, gmm::dense_matrix<double> &Klwlw)
{
    // Get shape functions and Gauss points
    Cache &surCacheUp = we->surUpE->getVCache();
    Gauss &surGaussUp = surCacheUp.getVGauss();
    Cache &surCacheLw = we->surLwE->getVCache();
    Gauss &surGaussLw = surCacheLw.getVGauss();
    Cache &volUpCache = we->volUpE->getVCache();
    Cache &volLwCache = we->volLwE->getVCache();
    // Get Jacobian
    Mem &surMemUp = we->surUpE->getVMem();
    Mem &surMemLw = we->surLwE->getVMem();
    Mem &volUpMem = we->volUpE->getVMem();
    Mem &volLwMem = we->volLwE->getVMem();
    // Reset matrices
    gmm::clear(Kupup);
    gmm::clear(Kuplw);
    gmm::clear(Klwup);
    gmm::clear(Klwlw);

    // invJ*dN
    gmm::dense_matrix<double> dNUp(we->nColUp, sol->pbl->nDim);
    gmm::dense_matrix<double> dNLw(we->nColLw, sol->pbl->nDim);
    for (size_t i = 0; i < we->nColUp; ++i)
    {
        for (size_t j = 0; j < sol->pbl->nDim; ++j)
        {
            dNUp(i, j) = volUpCache.dff[0][i][j];
        }
    }
    for (size_t i = 0; i < we->nColLw; ++i)
    {
        for (size_t j = 0; j < sol->pbl->nDim; ++j)
        {
            dNLw(i, j) = volLwCache.dff[0][i][j];
        }
    }
    gmm::dense_matrix<double> JdNUp(we->nColUp, sol->pbl->nDim);
    gmm::dense_matrix<double> JdNLw(we->nColLw, sol->pbl->nDim);
    gmm::mult(dNUp, gmm::transposed(volUpMem.getJinv(0)), JdNUp);
    gmm::mult(dNLw, gmm::transposed(volLwMem.getJinv(0)), JdNLw);
    // unit normal
    gmm::dense_matrix<double> nUp(sol->pbl->nDim, 1);
    gmm::dense_matrix<double> nLw(sol->pbl->nDim, 1);
    nUp(0, 0) = we->surUpE->normal().x[0];
    nUp(sol->pbl->nDim - 1, 0) = we->surUpE->normal().x[sol->pbl->nDim - 1];
    nLw(0, 0) = we->surLwE->normal().x[0];
    nLw(sol->pbl->nDim - 1, 0) = we->surLwE->normal().x[sol->pbl->nDim - 1];
    // velocity
    std::vector<double> gradUp(sol->pbl->nDim);
    std::vector<double> gradLw(sol->pbl->nDim);
    we->volUpE->computeGrad(sol->phi, gradUp, 0);
    we->volLwE->computeGrad(sol->phi, gradLw, 0);
    gmm::dense_matrix<double> Vup(sol->pbl->nDim, 1);
    gmm::dense_matrix<double> Vlw(sol->pbl->nDim, 1);
    for (size_t i = 0; i < sol->pbl->nDim; ++i)
    {
        Vup(i, 0) = gradUp[i];
        Vlw(i, 0) = gradLw[i];
    }
    // intermediate matrices
    gmm::dense_matrix<double> nJdNUp(we->nColUp, 1);
    gmm::dense_matrix<double> nJdNLw(we->nColLw, 1);
    gmm::dense_matrix<double> VJdNUp(we->nColUp, 1);
    gmm::dense_matrix<double> VJdNLw(we->nColLw, 1);
    gmm::mult(JdNUp, nUp, nJdNUp);
    gmm::mult(JdNLw, nLw, nJdNLw);
    gmm::mult(JdNUp, Vup, VJdNUp);
    gmm::mult(JdNLw, Vlw, VJdNLw);

    // Build
    /* @todo assemble on contributing nodes for lower surfaces => nRow <- we->surE->nodes.size() */
    for (size_t k = 0; k < surGaussUp.p.size(); ++k)
    {
        // N
        gmm::dense_matrix<double> N(1, we->nRow);
        for (size_t i = 0; i < we->nRow; ++i)
            N(0, i) = surCacheUp.ff[k][i];
        // Kn = N*VJdN
        gmm::dense_matrix<double> KnUp(we->nColUp, we->nRow);
        gmm::dense_matrix<double> KnLw(we->nColLw, we->nRow);
        gmm::mult(VJdNUp, N, KnUp);
        gmm::mult(VJdNLw, N, KnLw);
        // Kn = Kn*w*dtm
        gmm::scale(KnUp, surGaussUp.w[k] * surMemUp.getDetJ(k));
        gmm::scale(KnLw, surGaussUp.w[k] * surMemUp.getDetJ(k));
        // K = K + Kn
        gmm::add(KnUp, Kupup);
        gmm::add(KnLw, Klwup);
    }
    for (size_t k = 0; k < surGaussLw.p.size(); ++k)
    {
        // N
        gmm::dense_matrix<double> N(1, we->nRow);
        for (size_t i = 0; i < we->nRow; ++i)
            N(0, i) = surCacheLw.ff[k][i];
        // Kn = N*nJdN
        gmm::dense_matrix<double> KnUp(we->nColUp, we->nRow);
        gmm::dense_matrix<double> KnLw(we->nColLw, we->nRow);
        gmm::mult(nJdNUp, N, KnUp);
        gmm::mult(nJdNLw, N, KnLw);
        // Kn = Kn*w*dtm
        gmm::scale(KnUp, surGaussLw.w[k] * surMemLw.getDetJ(k));
        gmm::scale(KnLw, surGaussLw.w[k] * surMemLw.getDetJ(k));
        // K = K + Kn
        gmm::add(KnUp, Kuplw);
        gmm::add(KnLw, Klwlw);
    }
}

/**
 * @brief Write the results
 * @authors Adrien Crovato
 */
void Adjoint::save(int n, std::shared_ptr<MshExport> mshWriter)
{
    // Write files
    std::cout << "Saving files... " << std::endl;
    // setup results
    Results results;
    results.scalars_at_nodes["lambdaL"] = &lambdaL;
    results.scalars_at_nodes["lambdaD"] = &lambdaD;
    // save (all mesh and boundary surface)
    if (n > 0)
    {
        mshWriter->save(sol->pbl->msh->name + "_adjoint_" + std::to_string(n), results);
        for (auto sur : sol->pbl->bnd)
            sur->save(sur->groups[0]->tag->name + "adjoint_" + std::to_string(n), results);
    }
    else
    {
        mshWriter->save(sol->pbl->msh->name + "_adjoint", results);
        for (auto sur : sol->pbl->bnd)
            sur->save(sur->groups[0]->tag->name + "_adjoint", results);
    }
}

void Adjoint::write(std::ostream &out) const
{
    out << "flow::Adjoint"
        << "\n";
}
