/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wKuttaElement.h"
#include "wNode.h"
#include "wMem.h"
#include "wCache.h"
#include "wGauss.h"
using namespace tbox;
using namespace flow;

KuttaElement::KuttaElement(size_t _no, Element *&_surE, Element *&_volE,
                           std::vector<std::pair<size_t, Node *>> &_teN, int _sign)
    : wObject(), no(_no), surE(_surE), volE(_volE),
      teN(_teN), sign(_sign)

{
    // Sanity checks
    if (surE->type() == ELTYPE::LINE2)
        nDim = 2;
    else
    {
        std::stringstream err;
        err << "KuttaElement only implemented for surface elements of type "
            << ELTYPE::LINE2 << " (" << surE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    if (volE->type() == ELTYPE::TRI3)
        nCol = volE->nodes.size();
    else
    {
        std::stringstream err;
        err << "KuttaElement only implemented for volume elements of type "
            << ELTYPE::TRI3 << " (" << volE->type() << " was given)!\n";
        throw std::runtime_error(err.str());
    }
    // Count TE nodes
    nRow = teN.size();
}

/**
 * @brief Compute stiffness matrices associated to Kutta
 * 
 * K_ij = int V*dN_j * N_i dS
 *
 *              3
 *              o
 *             / \
 *            /   \
 *           /     \
 *     1,1' o-x---x-o 2,2'
 *            A   B
 *
 *     I = sum_j w'_j psi_j N'_j V_j invJ_j dN_j phi_j dtm(J')_j
 *       = psi sum_j (w'_j N'_j V_j invJ_j dN_j dtm(J')_j) phi
 *       = (psi)|1X2 (sum_j K_j)|2X3 (phi)|3X1
 *
 *     K = w'[N'1][Vx Vy]inv([dx/dxi dy/dxi])[dxiN1 dxiN2 dxiN3] dtm([dx/dxi' dy/dxi']^2)
 *           [N'2]          ([dx/det dy/det])[detN1 detN2 detN3]
 *
 *             phi1  phi2  phi3
 *     K = 1' [ .     .     .  ]
 *         2' [ .     .     .  ]
 *
 *     NB: K should be eval at GP A and B, for upper and lower element
 *     NB: V = grad(phi)
 * @authors Adrien Crovato
 */
void KuttaElement::buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, double alpha)
{
    // Get shape functions and Gauss points
    Cache &surCache = surE->getVCache();
    Gauss &surGauss = surCache.getVGauss();
    Cache &volCache = volE->getVCache();
    // Get Jacobian
    Mem &surMem = surE->getVMem();
    Mem &volMem = volE->getVMem();
    // Reset matrix
    gmm::clear(K);

    // Volume contribution
    // V
    std::vector<double> grad(nDim);
    volE->computeGrad(u, grad, 0);
    gmm::dense_matrix<double> V(1, nDim);
    for (size_t i = 0; i < nDim; ++i)
        V(0, i) = grad[i];
    // dN
    gmm::dense_matrix<double> dN(nDim, nCol);
    for (size_t j = 0; j < nCol; ++j)
    {
        for (size_t i = 0; i < nDim; ++i)
        {
            dN(i, j) = volCache.dff[0][j][i];
        }
    }
    // A = inv(J)*dN
    gmm::dense_matrix<double> A(nDim, nCol);
    gmm::mult(volMem.getJinv(0), dN, A);
    // B = V*[inv(J)*dN]
    gmm::dense_matrix<double> B(1, nCol);
    gmm::mult(V, A, B);

    // Build
    for (size_t k = 0; k < surGauss.p.size(); ++k)
    {
        // N
        gmm::dense_matrix<double> N(nRow, 1);
        for (size_t i = 0; i < nRow; ++i)
            N(i, 0) = surCache.ff[k][teN[i].first];

        // Kn = N*B
        gmm::dense_matrix<double> Kn(nRow, nCol);
        gmm::mult(N, B, Kn);
        // Kn = Kn*w*dtm
        gmm::scale(Kn, sign * surGauss.w[k] * surMem.getDetJ(k));
        // K = K + Kn
        gmm::add(Kn, K);
    }
}

/**
 * @brief Compute stiffness matrices associated to Kutta for Newton-Raphson
 * 
 * K_ij = int 2*V*dN_j * N_i dS
 * @authors Adrien Crovato
 */
void KuttaElement::buildNK(gmm::dense_matrix<double> &K, std::vector<double> const &u, double alpha)
{
    // Build
    buildK(K, u, alpha);
    gmm::scale(K, 2.);
}

/**
 * @brief Compute flux vectors associated to Kutta for Newton-Raphson
 * 
 * s_i = int V*V * N_i dS
 * @authors Adrien Crovato
 */
void KuttaElement::buildNs(std::vector<double> &b, std::vector<double> const &u, double alpha)
{
    // Get shape functions and Gauss points
    Cache &surCache = surE->getVCache();
    Gauss &surGauss = surCache.getVGauss();
    // Get Jacobian
    Mem &surMem = surE->getVMem();
    // Reset matrices
    gmm::clear(b);

    // Build
    for (size_t k = 0; k < surGauss.p.size(); ++k)
    {
        // N
        gmm::dense_matrix<double> N(nRow, 1);
        for (size_t i = 0; i < nRow; ++i)
            N(i, 0) = surCache.ff[k][teN[i].first];
        // V2
        double grad2 = volE->computeGrad2(u, 0);

        // Compute flux vector
        for (size_t i = 0; i < nRow; ++i)
            b[i] += sign * N(i, 0) * grad2 * surGauss.w[k] * surMem.getDetJ(k);
    }
}

void KuttaElement::write(std::ostream &out) const
{
    out << "KuttaElement #" << no
        << "\n\ton surface element (" << surE->type() << "): " << surE->no
        << "\n\ton volume element: (" << volE->type() << "): " << volE->no
        << "\n\twith " << teN.size() << " TE nodes: ";
    for (auto p : teN)
        out << p.second->no << " ";
    out << "\n\twith sign: " << sign << std::endl;
}