/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WKUTTAELEMENT_H
#define WKUTTAELEMENT_H

#include "flow.h"
#include "wObject.h"
#include "wElement.h"
#include <vector>

namespace flow
{

/**
 * @brief Kutta finite element
 * 
 * Made up of one surface element at the trailing edge and the connected volume
 * element. Only Line2 (surface) and Tri3 (volume) elements are curently
 * implemented.
 * Since Tri3 are used, the evaluation of the volume integrals is performed at
 * dummy Gauss point #0. If volume elements with non-constant shape function
 * derivatives are used, those derivates should be recomputed at the true
 * surface element Gauss points.
 * Kutta works well in 2D, but fails in 3D unless the geometry follows strict
 * restrictions (thin and sharp TE, simple planform, same mesh elements on TE
 * suction and pressure side, ...). Kutta contribution is therefore disabled for
 * 3D cases.
 * @authors Adrien Crovato
 */
class FLOW_API KuttaElement : public fwk::wObject
{
public:
    size_t no;           ///< ID
    tbox::Element *surE; ///< Surface element
    tbox::Element *volE; ///< Connected volume element

    size_t nRow;                                      ///< number of rows in stiffness matrix
    size_t nCol;                                      ///< number of columns in stiffness matrix
    std::vector<std::pair<size_t, tbox::Node *>> teN; ///< Map of local node indices and trailing edge nodes

    KuttaElement(size_t _no, tbox::Element *&_surE, tbox::Element *&_volE, std::vector<std::pair<size_t, tbox::Node *>> &_teN, int _sign);
    ~KuttaElement() {}

#ifndef SWIG
    void buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, double alpha);
    void buildNK(gmm::dense_matrix<double> &K, std::vector<double> const &u, double alpha);
    void buildNs(std::vector<double> &b, std::vector<double> const &u, double alpha);
    virtual void write(std::ostream &out) const override;
#endif

private:
    size_t nDim; ///< dimension of volume element
    int sign;    ///< +1 if upper element, -1 if lower element
};

} // namespace flow

#endif //WKUTTAELEMENT_H
