/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wProblem.h"
#include "wElement.h"
#include "wMem.h"
#include "wMedium.h"
#include "wBoundary.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wBlowing.h"
using namespace tbox;
using namespace flow;

Problem::Problem(std::shared_ptr<MshData> _msh, int dim, double aoa,
                 double minf, double sref, double cref, double xref,
                 double yref, double zref) : msh(_msh), nDim(dim), alpha(aoa),
                                             M_inf(minf), S_ref(sref), c_ref(cref)
{
    x_ref.x[0] = xref;
    x_ref.x[1] = yref;
    x_ref.x[2] = zref;
}

/**
 * @brief Set the fluid medium
 * @authors Adrien Crovato
 */
void Problem::set(std::shared_ptr<Medium> m)
{
    medium = m;
}

/**
 * @brief Add a boundary surface
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Boundary> b)
{
    bnd.push_back(b);
}

/**
 * @brief Add Dirichlet boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Dirichlet> d)
{
    dBCs.push_back(d);
}

/**
 * @brief Add initial condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Initial> i)
{
    iICs.push_back(i);
}

/**
 * @brief Add freestream (Neumann) boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Freestream> f)
{
    fBCs.push_back(f);
}

/**
 * @brief Add wake boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Wake> w)
{
    wBCs.push_back(w);
}

/**
 * @brief Add Kutta condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Kutta> k)
{
    kSCs.push_back(k);
}

/**
 * @brief Add Blowing boundary condition
 * @authors Adrien Crovato
 */
void Problem::add(std::shared_ptr<Blowing> b)
{
    bBCs.push_back(b);
}

/**
 * @brief Update the elements memory (Jacobian)
 * @authors Adrien Crovato
 */
void Problem::updateMem()
{
    // Update volume Jacobian
    for (auto e : medium->tag->elems)
        e->getVMem().update(true);
    // Update surface Jacobian (boundaries)
    for (auto surf : bnd)
        for (auto e : surf->groups[0]->tag->elems)
            e->getVMem().update(false);
    // Update surface Jacobian (Freestream B.C.)
    for (auto surf : fBCs)
        for (auto e : surf->tag->elems)
            e->getVMem().update(false);
    // Update surface Jacobian (Wake B.C.)
    for (auto surf : wBCs)
    {
        for (auto e : surf->groups[0]->tag->elems)
            e->getVMem().update(false);
        for (auto e : surf->groups[1]->tag->elems)
            e->getVMem().update(false);
    }
    // Update surface Jacobian (Blowing B.C.) if not already done
    for (auto surf : bBCs)
        for (auto e : surf->groups[0]->tag->elems)
            if (!(e->getVMem().valid))
                e->getVMem().update(false);
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 * @authors Adrien Crovato
 */
void Problem::check() const
{
    // Sanity checks
    if (medium == NULL)
        throw std::runtime_error("No Medium provided!\n");
    if (iICs.empty())
        throw std::runtime_error("No Initial Condition provided!\n");
    if (fBCs.empty() && dBCs.empty())
        throw std::runtime_error("No Freestream or Dirichlet B.C. provided!\n");
    // Three-dimension problem
    if (nDim == 3)
    {
        // Volume element type
        for (auto e : medium->tag->elems)
        {
            if (e->type() != ELTYPE::TETRA4)
            {
                std::stringstream err;
                err << "3D solver is only implemented for volume elements of type "
                    << ELTYPE::TETRA4 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
        }
        // Surface element type (Freestream B.C.)
        for (auto surf : fBCs)
        {
            for (auto e : surf->tag->elems)
            {
                if (e->type() != ELTYPE::TRI3)
                {
                    std::stringstream err;
                    err << "3D solver is only implemented for surface elements of type "
                        << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
        // Surface element type (Wake B.C.)
        for (auto surf : wBCs)
        {
            for (auto e : surf->groups[0]->tag->elems)
            {
                if (e->type() != ELTYPE::TRI3)
                {
                    std::stringstream err;
                    err << "3D solver is only implemented for surface elements of type "
                        << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
            for (auto e : surf->groups[1]->tag->elems)
            {
                if (e->type() != ELTYPE::TRI3)
                {
                    std::stringstream err;
                    err << "3D solver is only implemented for surface elements of type "
                        << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
        // Blowing B.C.
        if (!bBCs.empty())
            throw std::runtime_error("Blowing boundary conditions are not supported for 3D problems!\n");
    }
    // Two-dimension problem
    else if (nDim == 2)
    {
        // Volume element type
        for (auto e : medium->tag->elems)
        {
            if (e->type() != ELTYPE::TRI3)
            {
                std::stringstream err;
                err << "2D solver is only implemented for volume elements of type "
                    << ELTYPE::TRI3 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
        }
        // Surface element type (Freestream B.C.)
        for (auto surf : fBCs)
        {
            for (auto e : surf->tag->elems)
            {
                if (e->type() != ELTYPE::LINE2)
                {
                    std::stringstream err;
                    err << "2D solver is only implemented for surface elements of type "
                        << ELTYPE::LINE2 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
        // Surface element type (Wake B.C.)
        for (auto surf : wBCs)
        {
            for (auto e : surf->groups[0]->tag->elems)
            {
                if (e->type() != ELTYPE::LINE2)
                {
                    std::stringstream err;
                    err << "2D solver is only implemented for surface elements of type "
                        << ELTYPE::LINE2 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
            for (auto e : surf->groups[1]->tag->elems)
            {
                if (e->type() != ELTYPE::LINE2)
                {
                    std::stringstream err;
                    err << "2D solver is only implemented for surface elements of type "
                        << ELTYPE::LINE2 << " (" << e->type() << " was given)!\n";
                    throw std::runtime_error(err.str());
                }
            }
        }
    }
    else
    {
        std::stringstream err;
        err << "Solver is only implemented for 2D or 3D problems (dimension "
            << nDim << " was given)!\n";
        throw std::runtime_error(err.str());
    }
}

void Problem::write(std::ostream &out) const
{
    out << "flow::Problem parameters"
        << "\n\tDimension: " << nDim
        << "\n\tAngle of attack: " << alpha
        << "\n\tFreestream Mach number: " << M_inf
        << "\n\tReference area: " << S_ref
        << "\n\tReference length: " << c_ref
        << "\n\tReference point: [" << x_ref.x[0] << ", " << x_ref.x[1] << ", " << x_ref.x[2] << ']'
        << std::endl;
    out << "with"
        << "\n\t" << medium;
    for (auto b : bnd)
        out << "\n\t" << b;
    for (auto d : dBCs)
        out << "\n\t" << d;
    for (auto i : iICs)
        out << "\n\t" << i;
    for (auto f : fBCs)
        out << "\n\t" << f;
    for (auto w : wBCs)
        out << "\n\t" << w;
    for (auto k : kSCs)
        out << "\n\t" << k;
    for (auto b : bBCs)
        out << "\n\t" << b;
}
