#!/usr/bin/env python
# -*- coding: utf-8 -*-
# test multiscale

# mpiexec.openmpi --bycore --bind-to-core -n 6 ./run.py heat/tests/jalil2.py


from __future__ import print_function
from __future__ import division
from builtins import next
from builtins import range
from past.utils import old_div
from builtins import object
import heat as h
import tbox
import tbox.gmsh as gmsh
import math

try:
    import mpi4py.MPI as mpi
    comm = mpi.COMM_WORLD
    rank = comm.rank
    siz  = comm.size
    name = mpi.Get_processor_name()
    print("info: MPI found")
except:
    comm = None
    rank = 0
    siz  = 1
    name = "noname"
    print("info: MPI not found => MPI disabled")

class Job(object):
    """ Class containing data transmitted between MPI procs
    """
    def __init__(self, id, TM, dTxM, dTyM):
        # input
        self.id = id   # debug
        self.TM   = TM
        self.dTxM = dTxM
        self.dTyM = dTyM
        # output
        self.K11 = 0.0
        self.K22 = 0.0 # useless (=K11)
        self.K12 = 0.0
        self.K21 = 0.0 # useless (=-K12)
        # killme
        self.killme = False


class MicroModel(tbox.Fct2UdU):
    def __init__(self):
        """
        creation du micro modele
        """
        tbox.Fct2UdU.__init__(self)
        print("[m] creation of the micro model")

        self.msh = gmsh.MeshLoader("jalil2_honey.geo", __file__).execute()
        self.pbl = h.Problem(self.msh)
    
        self.solver = h.Solver(self.pbl)
        self.solver.nthreads = 1
        self.solver.save = False
        self.solver.verbose = 0
        
        self.fa=tbox.PwLf()
        self.fa.add(0, 80)
        self.fa.add(300, 80)
        self.fa.add(350, 20)
        self.fa.add(450, 10)
        self.fa.add(550, 5)
        self.fa.add(750, 1)
        self.fa.add(1000, 1)
        
        self.fs=tbox.PwLf()
        self.fs.add(0, 140)
        self.fs.add(300, 140)
        self.fs.add(350, 180)
        self.fs.add(450, 200)
        self.fs.add(550, 210)
        self.fs.add(750, 220)
        self.fs.add(1000, 220)
        
        self.fz = tbox.PwLf()
        self.fz.add(0,0) 
           
        h.Medium(self.pbl, "Air", tbox.Fct2PwLf(self.fa,self.fa,self.fz), 1.)
        h.Medium(self.pbl, "Skeleton", tbox.Fct2PwLf(self.fs,self.fs,self.fz), 1.)
        
        self.p = h.Periodic(self.pbl, 0.0, 0.0, 0.0)    
       
        self.verb = False
        
        self.jbz = {}      # liste des jobs
        
    def execute(self, job):

        TM   = job.TM
        dTxM = job.dTxM
        dTyM = job.dTyM 
    
        self.p.reset(TM, dTxM, dTyM)
        self.solver.start() 
        
        # resulting flux qM
        qMx = self.p.qM.x[0]
        qMy = self.p.qM.x[1]     
 
        if self.verb:
            print("[m] qMx =", qMx) 
            print("[m] qMy =", qMy) 
        
        # convert flux to an anisotropic conductivity "k(2,2)"
        dTnorm=math.sqrt(dTxM*dTxM+dTyM*dTyM)
        qnorm=math.sqrt(qMx*qMx+qMy*qMy)
        ampl=old_div(qnorm,dTnorm)
        
        ps=-(dTxM*qMx+dTyM*qMy)/dTnorm/qnorm
        if ps>1.0: ps=1.0
        if ps<-1.0: ps=-1.0
        
        #print ps
        ang=math.acos(ps)
        #print "ang=", ang, "ampl=", ampl
        #print "qMx=", qMx, "ampl*dT=", dTxM*ampl, "res=", qMx-dTxM*ampl
        cos=math.cos(ang)
        sin=math.sin(ang)
        job.K11 = cos*ampl
        job.K21 = sin*ampl
        job.K22 = job.K11   # => K22 inutile!
        job.K12 = -job.K21  # => K12 inutile!
       
    def eval(self, elem, npg, u, gradu, k, fake):  # tjs exécuté par rank#0
        if fake: # on cree un job
            TM   = u
            dTxM = gradu[0]
            dTyM = gradu[1]        
            dTnorm=math.sqrt(dTxM*dTxM+dTyM*dTyM)
            if dTnorm==0.0:
                dTxM = 1.0
                dTyM = 1.0
                dTnorm=math.sqrt(dTxM*dTxM+dTyM*dTyM)
                
            if self.verb: 
                print("[m] TM   =",TM)
                print("[m] dTxM =",dTxM)
                print("[m] dTyM =",dTyM)
            
            job = self.jbz.get((elem.no,npg))
            if job: # job existe deja, on le modifie
                job.TM=TM
                job.dTxM=dTxM
                job.dTyM=dTyM            
            else: # on ajoute ce job
                self.jbz[(elem.no,npg)] = Job((elem.no,npg), TM, dTxM, dTyM)
                #print "[m] creating job", (elem.no,npg)
                
                
        else: # on retourne le resultat du job 
            
            job = self.jbz.get((elem.no,npg))
            if not job:
                print("[m] job", (elem.no,npg), "not found!")
            #else:
            #    print "[m] job", (elem.no,npg), "OK!"
            k[0,0]=job.K11
            k[1,1]=job.K22 
            k[1,0]=job.K21
            k[0,1]=job.K12


    def evalall(self):   # tjs exécuté par rank#0
    
        #print "looping on %d jobs" % len(self.jobs)
        
        # version 1 proc (master travaille seul)
        if siz==1:
            #for job in self.jbz.itervalues():
            for id,job in self.jbz.items():
                self.execute(job) 
            return

        # version N procs
        it=self.jbz.__iter__()
        njobs=len(self.jbz)
        slaves=list(range(1,siz))
        while True:
            #send jobs to slaves
            jlist=[]
            for s in slaves:
                try:
                    jobk = next(it) # get next job
                except:
                    break
                job = self.jbz[jobk]
                jlist.append( (job,s) )
                if self.verb: 
                    print("[%d] sending job %s to %s.." % (rank,jobk,s))          
                comm.send(job, dest = s)
                
            if not jlist: # plus de job?
                break
                    
            #receive jobs
            for (job,s) in jlist:
                if self.verb: print("[%d] waiting job from %d.." % (rank,s))          
                job2 = comm.recv(source = s)
                if self.verb: print("[%d] job received from %d!" % (rank,s))
                job.K11=job2.K11
                job.K22=job2.K22
                job.K12=job2.K12
                job.K21=job2.K21                
       

def main():
   
    myk = MicroModel() # chaque process a son objet "modèle micro" (distinct)
    
    # faire en sorte que le rank0 charge le micro en 1er
    
    if rank==0:
        if myk.verb: print("[%d] démarrage du modèle macro" % rank)
        msh = gmsh.MeshLoader("jalil1_corner.geo", __file__).execute()
        pbl = h.Problem(msh)

        solver = h.Solver(pbl)
        solver.nthreads = 1

        h.Medium(pbl, "Volume", myk)
        
        h.Dirichlet(pbl,"Top BC", 400.)
        h.Dirichlet(pbl,"Right BC", 600.)

        solver.restol = 1e-6
        solver.start()

        # kill slaves:
        slaves=list(range(1,siz))
        for s in slaves:
            job=Job((0,0),0,0,0); job.killme=True
            if myk.verb: print("[%d] sending suicide job to %d" % (rank,s))
            comm.send(job, dest = s)

        from heat.viewer import GUI
        GUI().open('jalil1_corner')        
        
    else:
        myk.verb=False
        if myk.verb: print("[%d] démarrage d'un modèle micro" % rank)
               
        while True:
            # attend un gradient et cree un job
            if myk.verb: print("[%d] waiting for job from 0.." % rank)
            job = comm.recv(source = 0)  # recoit un job
            
            if job.killme:
                if myk.verb: print("[%d] Aaargl!" % rank)
                break
                
            if myk.verb: print("[%d] job received from 0!" % rank)
            myk.execute(job)             # execute le job
            if myk.verb: print("[%d] sending job results to 0.." % rank)
            comm.send(job, dest = 0)     # envoie le resultat
            if myk.verb: print("[%d] job results sent to 0.." % rank)

if __name__ == "__main__":
    main()
    print("[%d] end." % rank)
    
