/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wSolver.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTri3.h"
#include "wTimer.h"
#include "wTag.h"
#include "wFct2.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wSource.h"
#include "wBoundary.h"
#include "wDirichlet.h"
#include "wDisplayHook.h"
#include "wPeriodic.h"
#include "wResults.h"
#include "wMshExport.h"
#include "gmm_gmm.h"
#include <gmm/gmm_MUMPS_interface.h>
#include "gmm_extra.h"
#include "std_extra.h"
#include <typeinfo>
#include <algorithm>
#include <iomanip>

#include <tbb/task_scheduler_init.h>
#include <tbb/tbb.h>
#include <tbb/parallel_do.h>
#include <tbb/spin_mutex.h>
using namespace heat;
using namespace fwk;

#undef min
#undef max

Solver::Solver(std::shared_ptr<Problem> _pbl) : pbl(_pbl)
{
    dhook = NULL;

    // default values
    nthreads = 1;
    verbose = 1;
    stopit = false;
    restol = 1e-2;
    itmax = 100;
    save = true;

    // update element memory
    pbl->updateMem();
}

Solver::~Solver()
{
    std::cout << "~Solver()\n";
}

/**
 * @todo detecter/traiter les jacos negatifs pour tous les éléments
 * @todo planter clairement si pas de Medium
 * @todo calculer la taille du systeme en fct du nombre de noeuds des physical groups utilisés
 */
void Solver::start(std::shared_ptr<MshExport> mshWriter)
{
    tbb::spin_mutex mutex;
    tbb::task_scheduler_init init(nthreads);

    // start timers
    tms["total"].start();
    tms["prepro"].start(); // ----- start preprocessing phase

    if (verbose > 1)
        std::cout << *this;

    auto msh = pbl->msh;

    // assign a row for each node
    int i = 0;
    for (auto n : msh->nodes)
        n->row = i++;

    // setup results
    T1.resize(msh->nodes.size());
    gmm::clear(T1); // in case of restart

    Results results;
    results.scalars_at_nodes["T"] = &T1;

    // initial temperature guess
    if (!pbl->dBCs.empty())
    {
        double tmin = std::numeric_limits<double>::max();
        double tmax = std::numeric_limits<double>::min();
        for (auto dBC : pbl->dBCs)
        {
            double v = dBC->val;
            if (v < tmin)
                tmin = v;
            if (v > tmax)
                tmax = v;
        }
        /*
        std::cout << "tmin=" << tmin << "\n";
        std::cout << "tmax=" << tmax << "\n";
        */
        for (auto n : msh->nodes)
            T1[n->row] = (tmin + tmax) / 2;

        for (auto dBC : pbl->dBCs)
            dBC->apply(T1);
    }

    // periodic BCs
    if (pbl->pdic)
    {
        if (!pbl->dBCs.empty())
            throw std::runtime_error("cannot manage periodic BCs and Dirichlet BCs simultaneously!\n");
        pbl->pdic->init();

        // initial guess for T
        for (auto n : msh->nodes)
            T1[n->row] = pbl->pdic->TM;
    }
    tms["prepro"].stop(); // ----- end of preprocessing phase

    // --------------------------------------------------------------------
    // Iteration LOOP
    // --------------------------------------------------------------------

    int it = 0;
    while (it <= itmax)
    {
        std::vector<double> rhs(msh->nodes.size()); // RHS

        tms["Kass"].start();
        // K matrix assembly
        gmm::csr_matrix<double> K;
        buildK(K, rhs);
        //tomatlab("K.txt", K);
        //throw std::runtime_error("stop!");
        tms["Kass"].stop();

        tms["BCs"].start();

        if (!pbl->pdic)
        {
            builds(rhs); // sources
            buildq(rhs); // prescribed fluxes (=sources over lines)

            // BCs
            for (auto dBC : pbl->dBCs) // apply Dirichlet BCs to RHS
                dBC->apply(rhs);
        }
        tms["BCs"].stop();
        //std::cout << rhs << "\n";

        // residual calculation
        tms["residual"].start();

        // method #1

        std::vector<double> qint(msh->nodes.size());
        gmm::mult(K, T1, qint); // calcul de q_int = K T1:

        // method #2
        if (0)
        {
            std::vector<double> qint2(msh->nodes.size());
            buildqint(qint2); // calcul de q_int par integration du flux sur l'elem
            std::vector<double> diff(msh->nodes.size());
            gmm::add(qint, gmm::scaled(qint2, -1.0), diff);
            double normdiff = gmm::vect_norm2(diff);
            //if(verbose>1)
            std::cout << "norm diff=" << normdiff << '\n';
        }

        /*
        for(auto i=0; i<msh->nodes.size(); ++i)
            std::cout << i << ":" << qint[i] << "," << qint2[i] << "\n";
        */

        std::vector<double> res(msh->nodes.size()); // residual
        gmm::add(qint, gmm::scaled(rhs, -1.0), res);
        double rnorm2 = gmm::vect_norm2(res);
        double rnorminf = gmm::vect_norminf(res);
        tms["residual"].stop();

        if (verbose > 0)
        {
            std::cout << std::fixed << std::setprecision(3);
            std::cout << " * it #" << std::setw(3) << std::left << it
                      << " |r|2 = " << std::setw(9) << std::scientific << std::right << rnorm2
                      << "  |r|inf = " << std::setw(9) << std::scientific << std::right << rnorminf << '\n';
        }
        // check convergence
        if (rnorm2 < restol && it != 0)
            break;
        it++;

        // solve
        tms["solve"].start();
        if (verbose > 1)
            std::cout << "solving...\n";
        bool ok = gmm::MUMPS_solve(K, T1, rhs);
        if (!ok)
            throw std::runtime_error("problem while solving system with MUMPS");
        tms["solve"].stop();
    }
    if (it > itmax)
    {
        std::cout << "\n**WARNING! iterative process failed to converge\n";
        std::cout << "           in less than " << itmax << " iterations!\n\n"; // TODO: generer une erreur!
    }

    // --------------------------------------------------------------------
    // Postprocessing
    // --------------------------------------------------------------------

    if (verbose > 1)
        std::cout << "postpro...\n";
    tms["post"].start();

    // setup results

    std::vector<double> gradT_x(msh->elems.size()); ///< gradT_x
    std::vector<double> gradT_y(msh->elems.size()); ///< gradT_y
    std::vector<Pt> gradT(msh->elems.size());       ///< gradT
    std::vector<Pt> kgradT(msh->elems.size());      ///< gradT
    std::vector<double> kmoy11(msh->elems.size());  ///< mean k11 / element
    std::vector<double> kmoy22(msh->elems.size());  ///< mean k22 / element
    std::vector<double> kmoy12(msh->elems.size());  ///< mean k12 / element
    std::vector<double> kmoy21(msh->elems.size());  ///< mean k21 / element
    std::vector<gmm::dense_matrix<double>> kmoy(msh->elems.size());
    for (auto &km : kmoy)
        gmm::resize(km, 3, 3);

    results.scalars_at_elems["gradT_x"] = &gradT_x;
    results.scalars_at_elems["gradT_y"] = &gradT_y;
    results.scalars_at_elems["k11"] = &kmoy11;
    results.scalars_at_elems["k22"] = &kmoy22;
    results.scalars_at_elems["k12"] = &kmoy12;
    results.scalars_at_elems["k21"] = &kmoy21;
    results.vectors_at_elems["gradT"] = &gradT;
    results.vectors_at_elems["q [heat flux]"] = &kgradT;
    results.tensors_at_elems["k [tensor]"] = &kmoy;

    double Vtot = 0.0;
    std::vector<double> qM(2);
    for (auto mat : pbl->media)
    {
        tbb::parallel_do(mat->tag->elems.begin(), mat->tag->elems.end(),
                         [&](Element *e) {
                             if (e->type() != ELTYPE::TRI3)
                                 return;
                             //std::cout << "processing element #" << e->no << "\n";

                             // (flux moyen . volume) sur l'element
                             std::vector<double> qV(2);
                             e->computeqV(T1, qV, mat->k);
                             double V = e->computeV();

                             if (save)
                             {
                                 // gradient of T at GP 0
                                 std::vector<double> grad(2);
                                 e->computeGrad(T1, grad, 0); // at GP 0!
                                 int i = e->no - 1;
                                 gradT_x[i] = grad[0];
                                 gradT_y[i] = grad[1];
                                 gradT[i] = Pt(grad[0], grad[1]);

                                 // mean q over the element
                                 kgradT[i] = Pt(qV[0] / V, qV[1] / V);

                                 // mean k over the element
                                 gmm::dense_matrix<double> kmean(2, 2);
                                 e->computeV(T1, kmean, mat->k);
                                 gmm::scale(kmean, 1. / V);
                                 kmoy11[i] = kmean(0, 0);
                                 kmoy22[i] = kmean(1, 1);
                                 kmoy12[i] = kmean(0, 1);
                                 kmoy21[i] = kmean(1, 0);

                                 // tensor
                                 gmm::resize(kmean, 3, 3);
                                 gmm::copy(kmean, kmoy[i]);
                                 //std::cout << kmean << '\n';
                             }

                             tbb::spin_mutex::scoped_lock lock(mutex);
                             Vtot += V;
                             gmm::add(qV, qM); // qM = qM + qV
                         });
    }
    gmm::scale(qM, 1. / Vtot);

    if (verbose > 1)
    {
        std::cout << "Vtot = " << Vtot << '\n';
        std::cout << "qM = " << qM << '\n';
    }

    // periodic BCs
    if (pbl->pdic)
    {
        pbl->pdic->qM = Pt(qM[0], qM[1]);
        if (verbose > 1)
        {
            std::cout << "=> k_x = " << ((pbl->pdic->dTxM == 0.0) ? 0 : qM[0] / pbl->pdic->dTxM) << '\n';
            std::cout << "=> k_y = " << ((pbl->pdic->dTyM == 0.0) ? 0 : qM[1] / pbl->pdic->dTyM) << '\n';
        }
    }

    if (save)
    {
        if (verbose > 1)
            std::cout << "saving results to disk...\n";
        mshWriter->save(msh->name, results);
    }

    tms["post"].stop();
    tms["total"].stop();

    if (verbose > 0)
    {
        std::cout << "---CPU statistics ---\n";
        std::cout << tms;
        std::cout << "---------------------\n";
    }
}

void Solver::buildK(gmm::csr_matrix<double> &K2, std::vector<double> &rhs)
{
    tbb::spin_mutex mutex;

    std::shared_ptr<MshData> msh = pbl->msh;
    gmm::row_matrix<gmm::wsvector<double>> K(msh->nodes.size(), msh->nodes.size());

    // periodic BCs
    // (on va sommer les lignes des noeuds "d" sur celles des noeuds "n")

    std::vector<size_t> rows(msh->nodes.size());
    for (size_t i = 0; i < msh->nodes.size(); ++i)
        rows[i] = i;

    if (pbl->pdic)
    {
        for (auto np : pbl->pdic->BT)
            rows[np.d->row] = np.n->row;
        for (auto np : pbl->pdic->LR)
            rows[np.d->row] = np.n->row;
    }

    if (verbose > 1)
        std::cout << "building K (TBB/lambda) using " << nthreads << " threads...\n";

    bool useTBB = false;
    if (useTBB)
    {
        // 1-STEP PROCEDURE

        for (auto mat : pbl->media)
        {
            if (verbose > 1)
                std::cout << "\tprocessing " << *mat << '\n';
            tbb::parallel_do(mat->tag->elems.begin(), mat->tag->elems.end(),
                             [&](Element *e) {
                                 if (e->type() != ELTYPE::TRI3)
                                     return;
                                 //std::cout << "processing element #" << e->no << "\n";

                                 gmm::dense_matrix<double> Ke(3, 3);
                                 e->buildK(Ke, T1, mat->k, false);

                                 // assembly
                                 tbb::spin_mutex::scoped_lock lock(mutex);
                                 for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                                 {
                                     Node *nodi = e->nodes[ii];
                                     for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                                     {
                                         Node *nodj = e->nodes[jj];
                                         K(rows[nodi->row], nodj->row) += Ke(ii, jj);
                                     }
                                 }
                             });
        }
    }
    else
    {
        // 3-STEP PROCEDURE

        // *** build Job list - "fake run"

        if (verbose > 1)
            std::cout << "building job list...\n";

        for (auto mat : pbl->media)
        {
            if (verbose > 1)
                std::cout << "\tprocessing " << *mat << '\n';
            gmm::dense_matrix<double> Ke(3, 3); // bidon
            std::for_each(mat->tag->elems.begin(), mat->tag->elems.end(),
                          [&](Element *e) {
                              if (e->type() != ELTYPE::TRI3)
                                  return;
                              e->buildK(Ke, T1, mat->k, true);
                          });
        }

        // *** execute all jobs generated during the fake run

        if (verbose > 1)
            std::cout << "executing job list...\n";
        for (auto mat : pbl->media)
            mat->k.evalall();

        // ***  perform the assembly with the results ("true run")

        if (verbose > 1)
            std::cout << "performing assembly...\n";

        for (auto mat : pbl->media)
        {
            if (verbose > 1)
                std::cout << "\tprocessing " << *mat << '\n';

            std::for_each(mat->tag->elems.begin(), mat->tag->elems.end(), // devrait pouvoir etre fait en TBB...
                          [&](Element *e) {
                              if (e->type() != ELTYPE::TRI3)
                                  return;
                              gmm::dense_matrix<double> Ke(3, 3);
                              e->buildK(Ke, T1, mat->k, false);

                              // assembly
                              //tbb::spin_mutex::scoped_lock lock(mutex);
                              for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                              {
                                  Node *nodi = e->nodes[ii];
                                  for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                                  {
                                      Node *nodj = e->nodes[jj];
                                      K(rows[nodi->row], nodj->row) += Ke(ii, jj);
                                  }
                              }
                          });
        }
        if (verbose > 1)
            std::cout << "assembly done.\n";
    }

    // periodic BCs
    if (pbl->pdic)
    {
        // equations \theta_d (R,T) fct (L,B)
        double lx = pbl->pdic->corners[1]->pos.x[0] - pbl->pdic->corners[0]->pos.x[0];
        double ly = pbl->pdic->corners[3]->pos.x[1] - pbl->pdic->corners[0]->pos.x[1];
        if (verbose > 1)
            std::cout << "lx = " << lx << '\n';
        if (verbose > 1)
            std::cout << "ly = " << ly << '\n';
        double TM = pbl->pdic->TM;
        double dTxM = pbl->pdic->dTxM;
        double dTyM = pbl->pdic->dTyM;

        for (auto np : pbl->pdic->LR)
        {
            K.row(np.d->row) = gmm::wsvector<double>(msh->nodes.size()); // clears related row

            K(np.d->row, np.d->row) = 1.0;
            K(np.d->row, np.n->row) = -1.0;
            rhs[np.d->row] = lx * dTxM;
        }

        for (auto np : pbl->pdic->BT)
        {
            K.row(np.d->row) = gmm::wsvector<double>(msh->nodes.size()); // clears related row

            K(np.d->row, np.d->row) = 1.0;
            K(np.d->row, np.n->row) = -1.0;
            rhs[np.d->row] = ly * dTyM;
        }

        // corners \theta_2,3,4 fct \theta_1

        for (auto i = 0; i < 4; ++i)
        {
            Node *nod = pbl->pdic->corners[i];
            K.row(nod->row) = gmm::wsvector<double>(msh->nodes.size()); // clears related row
        }

        Node *n1 = pbl->pdic->corners[0];
        for (auto i = 1; i < 4; ++i)
        {
            Node *ni = pbl->pdic->corners[i];
            K(ni->row, ni->row) = 1.0;
            K(ni->row, n1->row) = -1.0;
        }
        Node *n2 = pbl->pdic->corners[1];
        Node *n3 = pbl->pdic->corners[2];
        Node *n4 = pbl->pdic->corners[3];
        rhs[n2->row] = lx * dTxM;
        rhs[n3->row] = lx * dTxM + ly * dTyM;
        rhs[n4->row] = ly * dTyM;

        // thermal energy consistency condition (n1->row)

        double Vtot = 0.0;
        double rhocM_Vtot = 0;
        for (auto mat : pbl->media)
        {
            //std::cout << "\tprocessing " << *mat << '\n';
            tbb::parallel_do(mat->tag->elems.begin(), mat->tag->elems.end(),
                             [&](Element *e) {
                                 if (e->type() != ELTYPE::TRI3)
                                     return;
                                 //std::cout << "processing element #" << e->no << "\n";

                                 std::vector<double> Ce(3);

                                 // ** Ce matrix => K matrix
                                 Fct0C f(1.0);
                                 e->builds(Ce, T1, f);

                                 // assembly
                                 tbb::spin_mutex::scoped_lock lock(mutex);
                                 for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                                 {
                                     Node *nodj = e->nodes[jj];
                                     K(n1->row, nodj->row) += (mat->rhoc) * Ce[jj];

                                     Vtot += Ce[jj];
                                     rhocM_Vtot += (mat->rhoc) * Ce[jj];
                                 }
                             });
        }
        rhs[n1->row] = rhocM_Vtot * TM;
        if (verbose > 1)
            std::cout << "Vtot=" << Vtot << "\n";
        if (verbose > 1)
            std::cout << "rhocM_Vtot=" << rhocM_Vtot << "\n";
    }

    if (!pbl->dBCs.empty())
    {
        if (verbose > 1)
            std::cout << "applying Dirichlet BCs...\n";

        // apply Dirichlet BCs to K
        for (auto dBC : pbl->dBCs)
        {
            for (auto nod : dBC->nodes)
            {
                //std::cout << "processing " << *nod << "\n";
                K.row(nod->row) = gmm::wsvector<double>(msh->nodes.size()); // clears related row
                K(nod->row, nod->row) = 1.0;
            }
        }
        if (verbose > 1)
            std::cout << "done." << std::endl;
    }

    // copy in a read-only CSR
    gmm::copy(K, K2);

    if (verbose > 1)
    {
        std::cout << "K (" << gmm::mat_nrows(K2) << "," << gmm::mat_ncols(K2) << ") nnz=" << gmm::nnz(K2) << "\n";
    }
    //gmm::cout << K2 << '\n';
}

void Solver::buildqint(std::vector<double> &qint)
{
    tbb::spin_mutex mutex;

    for (auto mat : pbl->media)
    {
        if (verbose > 1)
            std::cout << "\tprocessing " << *mat << '\n';
        tbb::parallel_do(mat->tag->elems.begin(), mat->tag->elems.end(),
                         [&](Element *e) {
                             if (e->type() != ELTYPE::TRI3)
                                 return;
                             //std::cout << "processing element #" << e->no << "\n";

                             std::vector<double> qe(e->nodes.size());
                             e->computeqint(T1, qe, mat->k);

                             // assembly
                             tbb::spin_mutex::scoped_lock lock(mutex);
                             for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                             {
                                 Node *nodi = e->nodes[ii];
                                 qint[nodi->row] += qe[ii];
                             }
                         });
    }

    if (!pbl->dBCs.empty())
    {
        if (verbose > 1)
            std::cout << "applying Dirichlet BCs...\n";

        // apply Dirichlet BCs to qint
        for (auto dBC : pbl->dBCs)
        {
            for (auto nod : dBC->nodes)
            {
                qint[nod->row] = T1[nod->row]; // pour comparer avec K.T1
            }
        }
        if (verbose > 1)
            std::cout << "done." << std::endl;
    }
}

/**
 * @brief volume sources
 */

void Solver::builds(std::vector<double> &s)
{
    tbb::spin_mutex mutex;

    if (verbose > 1)
        std::cout << "building s (TBB/lambda) using " << nthreads << " threads...\n";

    for (auto src : pbl->srcs)
    {
        //std::cout << "\tprocessing " << *src << '\n';
        tbb::parallel_do(src->tag->elems.begin(), src->tag->elems.end(),
                         [&](Element *e) {
                             if (e->type() != ELTYPE::TRI3 && e->type() != ELTYPE::LINE2)
                                 return;
                             //std::cout << "processing element #" << e->no << "\n";

                             std::vector<double> se(e->nodes.size());

                             // ** se vector => s vector
                             e->builds(se, T1, src->f);

                             // assembly
                             tbb::spin_mutex::scoped_lock lock(mutex);

                             for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                             {
                                 Node *nodi = e->nodes[ii];
                                 s[nodi->row] += se[ii];
                             }
                         });
    }

    if (verbose > 1)
    {
        std::cout << "s (" << gmm::vect_size(s) << ")\n";
    }
}

/**
 * @brief terme BC Neumann (pourrait etre mergé avec "builds"... c'est idem)
 * @todo merge this routine with builds
 */
void Solver::buildq(std::vector<double> &q)
{
    tbb::spin_mutex mutex;

    if (verbose > 1)
        std::cout << "building q (TBB/lambda) using " << nthreads << " threads...\n";

    for (auto bnd : pbl->bnds)
    {
        //std::cout << "\tprocessing " << *src << '\n';
        tbb::parallel_do(bnd->tag->elems.begin(), bnd->tag->elems.end(),
                         [&](Element *e) {
                             if (e->type() != ELTYPE::LINE2)
                                 return;
                             //std::cout << "processing element #" << e->no << "\n";

                             std::vector<double> qe(e->nodes.size());

                             // ** qe vector => q vector
                             e->builds(qe, T1, bnd->f);

                             // assembly
                             tbb::spin_mutex::scoped_lock lock(mutex);

                             for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                             {
                                 Node *nodi = e->nodes[ii];
                                 q[nodi->row] += qe[ii];
                             }
                         });
    }

    if (verbose > 1)
    {
        std::cout << "q (" << gmm::vect_size(q) << ")\n";
    }
}

void Solver::write(std::ostream &out) const
{
    out << "heat::Solver"
        << "\n";
}
