/**
 * @brief ElementVectors constructor
 * 
 * Arguments:
 * <ul>
 *  <li> domain: an RCP to a Domain object,
 *  <li> numPrimalDPN: the number of degrees of freedom per node (without taking into account the Lagrange multipliers),
 *  <li> pool_size: the number of threads that will be used to compute \f$\boldsymbol{N}\f$.
 * </ul>
 * 
 * The number of used threads is specify to allocate distinct memory for each thread to ensure that the threads
 * do not use the same memory addresses.
 */
template <typename scalar, int element_type, int element_size>
ElementVectors<scalar, element_type, element_size>::ElementVectors(Teuchos::RCP<katropton::Domain<scalar>> _domain,
                                                                   size_t _numPrimalDPN,
                                                                   size_t pool_size) : domain(_domain), numPrimalDPN(_numPrimalDPN)
{
    N = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A", pool_size, element_size, element_size);
}

/**
 * @brief Compute the \f$\boldsymbol{N}\f$ matrix.
 * Arguments:
 * <ul>
 *  <li> e: the local ID of the element for which \f$\boldsymbol{N}\f$ has to be computed,
 *  <li> i_thread: the ID of the thread used to compute \f$\boldsymbol{N}\f$.
 * </ul> 
 */
template <typename scalar, int element_type, int element_size>
void ElementVectors<scalar, element_type, element_size>::compute(int e, int i_thread)
{
    tbox::Cache *cache;

    std::vector<tbox::Pt> gauss_p;
    std::vector<double> gauss_w;

    /* @todo could probably be refactored into cache = element->getVCache(); */
    if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
    {
        tbox::CacheHex8 &cachehex8 = this->trilinosHex8GetCache();
        cache = &cachehex8;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
    {
        tbox::CacheTetra4 &cachetetra4 = this->trilinosTetra4GetCache();
        cache = &cachetetra4;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::TRI3))
    {
        tbox::CacheTri3 &cachetri3 = this->trilinosTri3GetCache();
        cache = &cachetri3;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }
    else if (element_type == static_cast<int>(tbox::ELTYPE::QUAD4))
    {
        tbox::CacheQuad4 &cachequad4 = this->trilinosQuad4GetCache();
        cache = &cachequad4;

        gauss_p = cache->getVGauss().p;
        gauss_w = cache->getVGauss().w;
    }

    tVector<double, element_size> ff;
    tMatrix<double, 3, 3> J;

    auto N_current = subview(N, i_thread, Kokkos::ALL(), Kokkos::ALL());

    Kokkos::deep_copy(N_current, 0);

    // Looping on gauss points
    for (auto a = 0; a < gauss_p.size(); ++a)
    {
        for (auto j = 0; j < element_size; ++j)
            if (element_type == static_cast<int>(tbox::ELTYPE::HEX8))
                ff(j) = (static_cast<tbox::CacheHex8 *>(cache)->ff)[a][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::TETRA4))
                ff(j) = (static_cast<tbox::CacheTetra4 *>(cache)->ff)[a][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::TRI3))
                ff(j) = (static_cast<tbox::CacheTri3 *>(cache)->ff)[a][j];
            else if (element_type == static_cast<int>(tbox::ELTYPE::QUAD4))
                ff(j) = (static_cast<tbox::CacheQuad4 *>(cache)->ff)[a][j];

        double detJ = this->buildJ(a, J, e, *(domain->elementsList), *(domain->nodesList));

        for (auto i = 0; i < element_size; ++i)
            for (auto j = 0; j < element_size; ++j)
                N_current(i, j) += ff(i) * ff(j) * gauss_w[a] * detJ;
    }
}

/**
 * @brief Return the \f$\boldsymbol{N}\f$.
 * 
 * Argument:
 * <ul>
 *  <li> i_thread: the ID of the thread used to compute \f$\boldsymbol{N}\f$.
 * </ul> 
 */
template <typename scalar, int element_type, int element_size>
Kokkos::View<scalar **, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace> ElementVectors<scalar, element_type, element_size>::getN(size_t i_thread) const
{
    Kokkos::View<scalar **, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace> N_current =
        subview(N, i_thread, Kokkos::ALL(), Kokkos::ALL());
    return N_current;
}