#include "Mortar.h"
//#include "wElement.h"

#include <MatrixMarket_Tpetra.hpp>

#include "Epetra_MpiComm.h"
#include "Epetra_SerialComm.h"
#include "Teuchos_ParameterList.hpp"

// MOERTEL headers
#include "mrtr_manager.H"
#include "mrtr_segment_bilinearquad.H"
#include "mrtr_segment_bilineartri.H"

/**
 * @brief Compute the Mortar matrices using Moertel and returns a Tpetra CrsMatrix
 * with double scalar type.
 */
Teuchos::RCP<Tpetra::CrsMatrix<>> compute_B(Teuchos::RCP<katropton::ElementsList> elementsList,
                                            Teuchos::RCP<katropton::NodesList> nodesList,
                                            Teuchos::RCP<katropton::ContactsList> contactsList,
                                            Teuchos::RCP<katropton::Map> map, MPI_Comm comm, bool SignoriniCheck)
{
    Epetra_MpiComm Comm(comm);
    Teuchos::RCP<const Teuchos::Comm<int>> Tcomm = Teuchos::rcp(new Teuchos::MpiComm<int>(comm));

    typedef katropton::Map::local_ordinal_type local_ordinal_type;
    typedef katropton::Map::global_ordinal_type global_ordinal_type;

    typedef Tpetra::CrsMatrix<> crs_matrix_type;

    const local_ordinal_type numMyContacts = contactsList->getContactNumber();

    // ------------------------------------------------------------- //
    // create numMyContacts empty MOERTEL::Interface instances
    // ------------------------------------------------------------- //
    int printlevel = 0;
    std::vector<Teuchos::RCP<MOERTEL::Interface>> interfaces(numMyContacts);
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i] = Teuchos::rcp(new MOERTEL::Interface(i, false, Comm, printlevel));

    // ------------------------------------------------------------- //
    // Add nodes on both sides of interface to interfaces
    // loop all nodes in the maps add them
    // to the interface with unique ids
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
    {
        const local_ordinal_type numMySlaveNodes = contactsList->getSlaveNodesSize(i);

        for (auto j = 0; j < numMySlaveNodes; ++j)
        {
            // get unique node id
            auto global_nodeid = contactsList->getSlaveNode(i, j);

            if (map->mapNodes->isNodeGlobalElement(global_nodeid))
            {
                // get node coordinates
                auto local_nodeid = map->mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = nodesList->nodes(local_nodeid, k);

                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, contactsList->isASlaveDirichletNode(i, global_nodeid), printlevel);
                // add node to interface i on side j
                interfaces[i]->AddNode(node, 0);
            }
        }

        const local_ordinal_type numMyMasterNodes = contactsList->getMasterNodesSize(i);

        for (auto j = 0; j < numMyMasterNodes; ++j)
        {
            // get unique node id
            auto global_nodeid = contactsList->getMasterNode(i, j);

            if (map->mapNodes->isNodeGlobalElement(global_nodeid))
            {
                // get node coordinates
                auto local_nodeid = map->mapNodesWO->getLocalElement(global_nodeid);
                double coord[3];
                for (auto k = 0; k < 3; ++k)
                    coord[k] = nodesList->nodes(local_nodeid, k);

                // create a moertel node
                MOERTEL::Node node(global_nodeid, coord, 1, &global_nodeid, false, printlevel);
                // add node to interface i on side j
                interfaces[i]->AddNode(node, 1);
            }
        }
    }

    // ------------------------------------------------------------- //
    // add segments on both sides of the interface to the interface
    // ------------------------------------------------------------- //

    auto i_index = 0;

    for (auto i = 0; i < numMyContacts; ++i)
    {
        const local_ordinal_type numMySlaveElems = contactsList->getSlaveElementsSize(i);

        for (auto j = 0; j < numMySlaveElems; ++j)
        {
            auto elem_local_index = contactsList->getSlaveElement(i, j);

            if (elementsList->getElementSize(elem_local_index) == 4)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[4];
                for (auto k = 0; k < 4; ++k)
                    nodeids[k] =
                        map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));

                MOERTEL::Segment_BiLinearQuad segment(i_index, 4, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 0);
            }
            else if (elementsList->getElementSize(elem_local_index) == 3)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[3];
                for (auto k = 0; k < 3; ++k)
                    nodeids[k] =
                        map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));
                MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 0);
            }
        }

        const local_ordinal_type numMyMasterElems = contactsList->getMasterElementsSize(i);

        for (auto j = 0; j < numMyMasterElems; ++j)
        {
            auto elem_local_index = contactsList->getMasterElement(i, j);

            if (elementsList->getElementSize(elem_local_index) == 4)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[4];
                for (auto k = 0; k < 4; ++k)
                    nodeids[k] =
                        map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));

                MOERTEL::Segment_BiLinearQuad segment(i_index, 4, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 1);
            }
            else if (elementsList->getElementSize(elem_local_index) == 3)
            {
                i_index = map->mapElems->getGlobalElement(elem_local_index);

                global_ordinal_type nodeids[3];
                for (auto k = 0; k < 3; ++k)
                    nodeids[k] =
                        map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));
                MOERTEL::Segment_BiLinearTri segment(i_index++, 3, nodeids, printlevel);
                interfaces[i]->AddSegment(segment, 1);
            }
        }
    }

    // ------------------------------------------------------------- //
    // choose the mortar side of the interface (0 or 1)
    //
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        interfaces[i]->SetMortarSide(1);

    // ------------------------------------------------------------- //
    // As we do not know the mortar side yet (we decided to le the
    // package choose it), we can not set a dual trace function (mortar space)
    // as we don't know the side to set it to
    // so we just give orders for the function type
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
    {
        if (contactsList->getSlaveElementsSize(i) >= 1)
        {
            if (elementsList->getElementSize(contactsList->getSlaveElement(i, 0)) == 4)
            {
                if (contactsList->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_BiLinearQuad);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_DualBiLinearQuad);
                }
            }
            else if (elementsList->getElementSize(contactsList->getSlaveElement(i, 0)) == 3)
            {
                if (contactsList->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_LinearTri);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_DualLinearTri);
                }
            }
        }

        else if (contactsList->getMasterElementsSize(i) >= 1)
        {
            if (elementsList->getElementSize(contactsList->getMasterElement(i, 0)) == 4)
            {
                if (contactsList->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_BiLinearQuad);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_BiLinearQuad,
                                                    MOERTEL::Function::func_DualBiLinearQuad);
                }
            }
            else if (elementsList->getElementSize(contactsList->getMasterElement(i, 0)) == 3)
            {
                if (contactsList->isStandardShapeFunction(i))
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_LinearTri);
                }
                else
                {
                    interfaces[i]->SetFunctionTypes(MOERTEL::Function::func_LinearTri,
                                                    MOERTEL::Function::func_DualLinearTri);
                }
            }
        }
    }
    // ------------------------------------------------------------- //
    // complete the interfaces
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        if (!interfaces[i]->Complete())
        {
            exit(EXIT_FAILURE);
        }
    // ------------------------------------------------------------- //
    // create an empty MOERTEL::Manager for 3D problems
    // It organizes everything from integration to solution
    // ------------------------------------------------------------- //
    MOERTEL::Manager manager(Comm, MOERTEL::Manager::manager_3D, printlevel);

    // ------------------------------------------------------------- //
    // Add the interfaces to the manager
    // ------------------------------------------------------------- //
    for (auto i = 0; i < numMyContacts; ++i)
        manager.AddInterface(*(interfaces[i]));

    // ------------------------------------------------------------- //
    // for mortar integration, the mortar manager needs to know about
    // the rowmap of the original (uncoupled) problem because it will
    // create coupling matrices D and M matching that rowmap
    // ------------------------------------------------------------- //
    Teuchos::ArrayView<const global_ordinal_type> myNodes = map->mapNodes->getNodeElementList();

    int numMyNodes = map->mapNodes->getNodeNumElements();
    int numNodes = map->mapNodes->getGlobalNumElements();

    int myNodesArray[numMyNodes];

    for (auto i = 0; i < numMyNodes; ++i)
        myNodesArray[i] = (int)myNodes[i];

    Epetra_Map map00((int)numNodes, (int)numMyNodes, myNodesArray, (int)map->mapNodes->getIndexBase(), Comm);
    manager.SetProblemMap(&map00);

    // ============================================================= //
    // choose integration parameters
    // ============================================================= //
    Teuchos::ParameterList &moertelparams = manager.Default_Parameters();
    // this does affect this 3D case only
    moertelparams.set("exact values at gauss points", true);
    // 1D interface possible values are 1,2,3,4,5,6,7,8,10 (2 recommended with linear shape functions)
    moertelparams.set("number gaussian points 1D", 2);
    // 2D interface possible values are 3,6,12,13,16,19,27 (12 recommended with linear functions)
    moertelparams.set("number gaussian points 2D", 12);

    // ============================================================= //
    // Here we are done with the construction phase of the interface
    // so we can integrate the mortar integrals
    // (Note we have not yet evaluated the PDE at all!)
    // ============================================================= //

    bool do_integrate = true;
    int n_integration = 0;
    int n_integration_max = 10;

    while (do_integrate)
    {
        try
        {
            do_integrate = false;
            ++n_integration;
            manager.Mortar_Integrate();
        }
        catch (...)
        {
            do_integrate = true;
        }

        if (n_integration >= n_integration_max)
            break;
    }

    if (n_integration != 1)
        std::cout << "Has integrate " << n_integration << " times" << std::endl;

    const Epetra_CrsMatrix *D = manager.D();
    const Epetra_CrsMatrix *M = manager.M();

    std::map<int, int> lm_to_dof = manager.lm_to_dof();

    const size_t numDPN = map->numPrimalDPN;
    const size_t numMechanicalLagrangeMultipliers = (D->NumGlobalCols() - numNodes);

    size_t numLMPN;
    if (contactsList->hasAtLeastOneSticking() && numDPN == 4)
        numLMPN = 4;
    else if (contactsList->hasAtLeastOneSticking())
        numLMPN = 3;
    else if (numDPN == 4)
        numLMPN = 2;
    else
        numLMPN = 1;

    map->numDualDPN = numLMPN;

    const size_t numTotalLagrangeMultipliers = numLMPN * numMechanicalLagrangeMultipliers;

    int indexBase = 0;

    // ---------------

    {
        Teuchos::ArrayView<const global_ordinal_type> myNodes = map->mapNodes->getNodeElementList();

        std::vector<global_ordinal_type> myLagrangeDofs = {};

        size_t numMyRows = D->NumMyRows();
        for (auto i = 0; i < numMyRows; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;
            size_t lagrange_id = global_id - numNodes;
            for (auto j = 0; j < numLMPN; ++j)
                myLagrangeDofs.push_back(lagrange_id * numLMPN + j);
        }

        map->mapLagrangeDofs = rcp(new Tpetra::Map<>(numTotalLagrangeMultipliers, myLagrangeDofs, indexBase, Tcomm));
    }

    // ---------------
    for (auto i = lm_to_dof.begin(); i != lm_to_dof.end(); ++i)
        map->lm_to_dof_global[i->first - numNodes] = i->second;

    for (auto i = lm_to_dof.begin(); i != lm_to_dof.end(); ++i)
        if (map->mapNodes->isNodeGlobalElement(i->second))
        {
            map->lm_to_dof[i->first - numNodes] = i->second;
        }

    for (auto i = map->lm_to_dof.begin(); i != map->lm_to_dof.end(); ++i)
        map->lagrange_to_dof[i->first - numNodes] = (i->second) * map->numPrimalDPN;

    // As lm_to_dof is one to one, we can create dof_to_lm as follows
    for (auto i = map->lm_to_dof.begin(); i != map->lm_to_dof.end(); ++i)
        map->dof_to_lm[i->second] = i->first;

    // ---------------

    // Compute the outward normal
    typedef Tpetra::MultiVector<double, local_ordinal_type, global_ordinal_type> double_multivector_type;

    Teuchos::RCP<double_multivector_type> normalWO = Teuchos::rcp(new double_multivector_type(map->mapNodesWO, 3 * numMyContacts, true));
    Teuchos::RCP<double_multivector_type> normalSUM = Teuchos::rcp(new double_multivector_type(map->mapNodes, 3 * numMyContacts, true));

    std::vector<global_ordinal_type> myNodes2 = {};

    for (auto i = 0; i < numNodes; ++i)
        myNodes2.push_back(i);

    Teuchos::RCP<Tpetra::Map<>> fullNodeMap = rcp(new Tpetra::Map<>(numNodes, myNodes2, indexBase, Tcomm));
    Teuchos::RCP<double_multivector_type> normal = Teuchos::rcp(new double_multivector_type(fullNodeMap, 3 * numMyContacts, true));

    {
        for (auto i = 0; i < numMyContacts; ++i)
        {
            const local_ordinal_type numMySlaveElems = contactsList->getSlaveElementsSize(i);

            for (auto j = 0; j < numMySlaveElems; ++j)
            {
                auto elem_local_index = contactsList->getSlaveElement(i, j);

                if (elementsList->getElementSize(elem_local_index) == 4)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[4];
                    double coordinates[4 * 3];
                    for (auto k = 0; k < 4; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodesList->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double vec3[3];
                    double current_normal[3];
                    double current_normal2[3];
                    double norm1;
                    double norm2;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    vec3[0] = coordinates[9] - coordinates[0];
                    vec3[1] = coordinates[10] - coordinates[1];
                    vec3[2] = coordinates[11] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    current_normal2[0] = vec2[1] * vec3[2] - vec2[2] * vec3[1];
                    current_normal2[1] = vec2[2] * vec3[0] - vec2[0] * vec3[2];
                    current_normal2[2] = vec2[0] * vec3[1] - vec2[1] * vec3[0];

                    norm1 = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));
                    norm2 = sqrt(pow(current_normal2[0], 2) + pow(current_normal2[1], 2) + pow(current_normal2[2], 2));

                    if (norm1 > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm1;
                        current_normal[1] = current_normal[1] / norm1;
                        current_normal[2] = current_normal[2] / norm1;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    current_area = (norm1 + norm2) / 2.;

                    for (auto k = 0; k < 4; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
                else if (elementsList->getElementSize(elem_local_index) == 3)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[3];
                    double coordinates[3 * 3];
                    for (auto k = 0; k < 3; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodesList->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double current_normal[3];
                    double norm;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                    current_area = norm / 2.;

                    if (norm > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm;
                        current_normal[1] = current_normal[1] / norm;
                        current_normal[2] = current_normal[2] / norm;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    for (auto k = 0; k < 3; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
            }

            const local_ordinal_type numMyMasterElems = contactsList->getMasterElementsSize(i);

            for (auto j = 0; j < numMyMasterElems; ++j)
            {
                auto elem_local_index = contactsList->getMasterElement(i, j);

                if (elementsList->getElementSize(elem_local_index) == 4)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[4];
                    double coordinates[4 * 3];
                    for (auto k = 0; k < 4; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodesList->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double vec3[3];
                    double current_normal[3];
                    double current_normal2[3];
                    double norm1;
                    double norm2;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    vec3[0] = coordinates[9] - coordinates[0];
                    vec3[1] = coordinates[10] - coordinates[1];
                    vec3[2] = coordinates[11] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    current_normal2[0] = vec2[1] * vec3[2] - vec2[2] * vec3[1];
                    current_normal2[1] = vec2[2] * vec3[0] - vec2[0] * vec3[2];
                    current_normal2[2] = vec2[0] * vec3[1] - vec2[1] * vec3[0];

                    norm1 = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));
                    norm2 = sqrt(pow(current_normal2[0], 2) + pow(current_normal2[1], 2) + pow(current_normal2[2], 2));

                    if (norm1 > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm1;
                        current_normal[1] = current_normal[1] / norm1;
                        current_normal[2] = current_normal[2] / norm1;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    current_area = (norm1 + norm2) / 2.;

                    for (auto k = 0; k < 4; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
                else if (elementsList->getElementSize(elem_local_index) == 3)
                {
                    i_index = map->mapElems->getGlobalElement(elem_local_index);

                    global_ordinal_type node_id[3];
                    double coordinates[3 * 3];
                    for (auto k = 0; k < 3; ++k)
                    {
                        node_id[k] = map->mapNodesWO->getGlobalElement(elementsList->getElementNode(elem_local_index, k));
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            coordinates[k * 3 + l] = nodesList->nodes(local_nodeid, l);
                    }

                    double vec1[3];
                    double vec2[3];
                    double current_normal[3];
                    double norm;
                    double current_area;

                    vec1[0] = coordinates[3] - coordinates[0];
                    vec1[1] = coordinates[4] - coordinates[1];
                    vec1[2] = coordinates[5] - coordinates[2];

                    vec2[0] = coordinates[6] - coordinates[0];
                    vec2[1] = coordinates[7] - coordinates[1];
                    vec2[2] = coordinates[8] - coordinates[2];

                    current_normal[0] = vec1[1] * vec2[2] - vec1[2] * vec2[1];
                    current_normal[1] = vec1[2] * vec2[0] - vec1[0] * vec2[2];
                    current_normal[2] = vec1[0] * vec2[1] - vec1[1] * vec2[0];

                    norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                    current_area = norm / 2.;

                    if (norm > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm;
                        current_normal[1] = current_normal[1] / norm;
                        current_normal[2] = current_normal[2] / norm;

                        for (auto l = 0; l < 3; ++l)
                            if (abs(current_normal[l]) <= 1e-10)
                                current_normal[l] = 0.;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    for (auto k = 0; k < 3; ++k)
                    {
                        auto local_nodeid = map->mapNodesWO->getLocalElement(node_id[k]);

                        for (auto l = 0; l < 3; ++l)
                            normalWO->sumIntoLocalValue(local_nodeid, i * 3 + l, current_area * current_normal[l]);
                    }
                }
            }
        }
    }
    Tpetra::Export<> export1(map->mapNodesWO, map->mapNodes);
    Tpetra::Export<> export2(map->mapNodes, normal->getMap());
    Tpetra::Import<> import1(map->mapNodes, normal->getMap());

    normalSUM->doExport(*normalWO, export1, Tpetra::ADD);
    normal->doImport(*normalSUM, import1, Tpetra::INSERT);

    normal->template sync<Kokkos::HostSpace>();
    auto normal_2d = normal->template getLocalView<Kokkos::HostSpace>();

    size_t maxNumEntPerRow = 3 * (D->MaxNumEntries() + M->MaxNumEntries());
    Teuchos::RCP<crs_matrix_type> B(new crs_matrix_type(map->mapLagrangeDofs, maxNumEntPerRow));
    {
        int DNumMyRow = D->NumMyRows();
        int DMaxNumIndices = D->MaxNumEntries();
        double *values = new double[DMaxNumIndices];
        int DNumIndices;
        int *Indices_int = new int[DMaxNumIndices];

        const Epetra_Map DrowMap = D->RowMap();
        const Epetra_Map DcolMap = D->ColMap();

        for (size_t i = 0; i < DNumMyRow; ++i)
        {
            size_t global_id = D->GRID(i);
            if (global_id < numNodes)
                continue;

            D->ExtractGlobalRowCopy(global_id, DMaxNumIndices, DNumIndices, values, Indices_int);

            // find which interface interface_j is related to the current Lagrange multiplier
            size_t interface_j = 0; // 0 by default
            if (contactsList->getContactNumber() > 1)
            {
                bool doIStop = false;
                for (auto k = 0; k < DNumIndices; ++k)
                {
                    for (auto j = 0; j < numMyContacts; ++j)
                    {
                        const local_ordinal_type numMySlaveNodes = contactsList->getSlaveNodesSize(j);

                        for (auto l = 0; l < numMySlaveNodes; ++l)
                        {
                            auto global_nodeid = contactsList->getSlaveNode(j, l);

                            if (global_nodeid == Indices_int[k])
                            {
                                interface_j = j;
                                doIStop = true;
                            }
                            if (doIStop)
                                break;
                        }
                        if (doIStop)
                            break;
                    }
                    if (doIStop)
                        break;
                }
            }

            global_ordinal_type row = (global_id - numNodes) * numLMPN;

            for (auto k = 0; k < DNumIndices; ++k)
            {
                global_ordinal_type col1 = Indices_int[k] * numDPN;
                global_ordinal_type col2 = Indices_int[k] * numDPN + 1;
                global_ordinal_type col3 = Indices_int[k] * numDPN + 2;

                global_ordinal_type global_node_id = Indices_int[k];
                local_ordinal_type local_node_id = normal->getMap()->getLocalElement(global_node_id);

                double current_normal[3];
                current_normal[0] = normal_2d(local_node_id, interface_j * 3 + 0);
                current_normal[1] = normal_2d(local_node_id, interface_j * 3 + 1);
                current_normal[2] = normal_2d(local_node_id, interface_j * 3 + 2);

                double norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                if (norm > 0.)
                {
                    current_normal[0] = current_normal[0] / norm;
                    current_normal[1] = current_normal[1] / norm;
                    current_normal[2] = current_normal[2] / norm;
                }
                else
                {
                    current_normal[0] = 0.;
                    current_normal[1] = 0.;
                    current_normal[2] = 0.;
                }

                double current_tangential_1[3];
                double current_tangential_2[3];

                current_tangential_1[0] = 0.;
                current_tangential_1[1] = 0.;
                current_tangential_1[2] = 1.;

                current_tangential_2[0] = 0.;
                current_tangential_2[1] = 0.;
                current_tangential_2[2] = 0.;

                if (contactsList->hasAtLeastOneSticking())
                {
                    double dot_n_z = current_normal[0] * current_tangential_1[0] + current_normal[1] * current_tangential_1[1] + current_normal[2] * current_tangential_1[2];

                    if (acos(abs(dot_n_z)) <= 1e-6 || contactsList->isTying(interface_j))
                    {
                        if (contactsList->isTying(interface_j))
                        {
                            current_normal[0] = 0.;
                            current_normal[1] = 0.;
                            current_normal[2] = 1.;
                        }
                        current_tangential_1[0] = 1.;
                        current_tangential_1[1] = 0.;
                        current_tangential_1[2] = 0.;

                        current_tangential_2[0] = 0.;
                        current_tangential_2[1] = 1.;
                        current_tangential_2[2] = 0.;
                    }
                    else
                    {
                        current_tangential_1[0] = current_tangential_1[0] - dot_n_z * current_normal[0];
                        current_tangential_1[1] = current_tangential_1[1] - dot_n_z * current_normal[1];
                        current_tangential_1[2] = current_tangential_1[2] - dot_n_z * current_normal[2];

                        double norm = sqrt(pow(current_tangential_1[0], 2) + pow(current_tangential_1[1], 2) + pow(current_tangential_1[2], 2));

                        current_tangential_1[0] = current_tangential_1[0] / norm;
                        current_tangential_1[1] = current_tangential_1[1] / norm;
                        current_tangential_1[2] = current_tangential_1[2] / norm;

                        current_tangential_2[0] = -(current_normal[1] * current_tangential_1[2] - current_normal[2] * current_tangential_1[1]);
                        current_tangential_2[1] = -(current_normal[2] * current_tangential_1[0] - current_normal[0] * current_tangential_1[2]);
                        current_tangential_2[2] = -(current_normal[0] * current_tangential_1[1] - current_normal[1] * current_tangential_1[0]);
                    }
                }

                double value1 = values[k] * current_normal[0] * contactsList->getMortarScale(interface_j);
                double value2 = values[k] * current_normal[1] * contactsList->getMortarScale(interface_j);
                double value3 = values[k] * current_normal[2] * contactsList->getMortarScale(interface_j);

                B->insertGlobalValues(row,
                                      Teuchos::tuple<global_ordinal_type>(col1),
                                      Teuchos::tuple<double>(value1));
                B->insertGlobalValues(row,
                                      Teuchos::tuple<global_ordinal_type>(col2),
                                      Teuchos::tuple<double>(value2));
                B->insertGlobalValues(row,
                                      Teuchos::tuple<global_ordinal_type>(col3),
                                      Teuchos::tuple<double>(value3));

                if (contactsList->isSticking(interface_j))
                {
                    value1 = values[k] * current_tangential_1[0] * contactsList->getMortarScale(interface_j);
                    value2 = values[k] * current_tangential_1[1] * contactsList->getMortarScale(interface_j);
                    value3 = values[k] * current_tangential_1[2] * contactsList->getMortarScale(interface_j);

                    B->insertGlobalValues(row + 1,
                                          Teuchos::tuple<global_ordinal_type>(col1),
                                          Teuchos::tuple<double>(value1));
                    B->insertGlobalValues(row + 1,
                                          Teuchos::tuple<global_ordinal_type>(col2),
                                          Teuchos::tuple<double>(value2));
                    B->insertGlobalValues(row + 1,
                                          Teuchos::tuple<global_ordinal_type>(col3),
                                          Teuchos::tuple<double>(value3));

                    value1 = values[k] * current_tangential_2[0] * contactsList->getMortarScale(interface_j);
                    value2 = values[k] * current_tangential_2[1] * contactsList->getMortarScale(interface_j);
                    value3 = values[k] * current_tangential_2[2] * contactsList->getMortarScale(interface_j);

                    B->insertGlobalValues(row + 2,
                                          Teuchos::tuple<global_ordinal_type>(col1),
                                          Teuchos::tuple<double>(value1));
                    B->insertGlobalValues(row + 2,
                                          Teuchos::tuple<global_ordinal_type>(col2),
                                          Teuchos::tuple<double>(value2));
                    B->insertGlobalValues(row + 2,
                                          Teuchos::tuple<global_ordinal_type>(col3),
                                          Teuchos::tuple<double>(value3));
                }
                if (numDPN == 4)
                {
                    global_ordinal_type row2 = row + numLMPN - 1;
                    global_ordinal_type col4 = Indices_int[k] * numDPN + 3;
                    B->insertGlobalValues(row2,
                                          Teuchos::tuple<global_ordinal_type>(col4),
                                          Teuchos::tuple<double>(values[k]));
                }
            }

            if (SignoriniCheck)
                if (contactsList->isNotSignorini(interface_j) == false)
                    continue;
            {
                int MMaxNumIndices = M->MaxNumEntries();
                double *values = new double[MMaxNumIndices];
                int MNumIndices;
                int *Indices_int = new int[MMaxNumIndices];

                M->ExtractGlobalRowCopy(global_id, MMaxNumIndices, MNumIndices, values, Indices_int);

                for (auto k = 0; k < MNumIndices; ++k)
                {
                    global_ordinal_type col1 = Indices_int[k] * numDPN;
                    global_ordinal_type col2 = Indices_int[k] * numDPN + 1;
                    global_ordinal_type col3 = Indices_int[k] * numDPN + 2;

                    global_ordinal_type global_node_id = Indices_int[k];
                    local_ordinal_type local_node_id = normal->getMap()->getLocalElement(global_node_id);

                    double current_normal[3];
                    current_normal[0] = normal_2d(local_node_id, interface_j * 3 + 0);
                    current_normal[1] = normal_2d(local_node_id, interface_j * 3 + 1);
                    current_normal[2] = normal_2d(local_node_id, interface_j * 3 + 2);

                    double norm = sqrt(pow(current_normal[0], 2) + pow(current_normal[1], 2) + pow(current_normal[2], 2));

                    if (norm > 0.)
                    {
                        current_normal[0] = current_normal[0] / norm;
                        current_normal[1] = current_normal[1] / norm;
                        current_normal[2] = current_normal[2] / norm;
                    }
                    else
                    {
                        current_normal[0] = 0.;
                        current_normal[1] = 0.;
                        current_normal[2] = 0.;
                    }

                    double current_tangential_1[3];
                    double current_tangential_2[3];

                    current_tangential_1[0] = 0.;
                    current_tangential_1[1] = 0.;
                    current_tangential_1[2] = 1.;

                    current_tangential_2[0] = 0.;
                    current_tangential_2[1] = 0.;
                    current_tangential_2[2] = 0.;

                    if (contactsList->hasAtLeastOneSticking())
                    {
                        current_tangential_1[0] = 0.;
                        current_tangential_1[1] = 0.;
                        current_tangential_1[2] = 1.;

                        double dot_n_z = current_normal[0] * current_tangential_1[0] + current_normal[1] * current_tangential_1[1] + current_normal[2] * current_tangential_1[2];

                        if (acos(abs(dot_n_z)) <= 1e-6 || contactsList->isTying(interface_j))
                        {
                            if (contactsList->isTying(interface_j))
                            {
                                current_normal[0] = 0.;
                                current_normal[1] = 0.;
                                current_normal[2] = -1.;
                            }
                            current_tangential_1[0] = 1.;
                            current_tangential_1[1] = 0.;
                            current_tangential_1[2] = 0.;

                            current_tangential_2[0] = 0.;
                            current_tangential_2[1] = 1.;
                            current_tangential_2[2] = 0.;
                        }
                        else
                        {
                            current_tangential_1[0] = current_tangential_1[0] - dot_n_z * current_normal[0];
                            current_tangential_1[1] = current_tangential_1[1] - dot_n_z * current_normal[1];
                            current_tangential_1[2] = current_tangential_1[2] - dot_n_z * current_normal[2];

                            double norm = sqrt(pow(current_tangential_1[0], 2) + pow(current_tangential_1[1], 2) + pow(current_tangential_1[2], 2));

                            current_tangential_1[0] = current_tangential_1[0] / norm;
                            current_tangential_1[1] = current_tangential_1[1] / norm;
                            current_tangential_1[2] = current_tangential_1[2] / norm;

                            current_tangential_2[0] = current_normal[1] * current_tangential_1[2] - current_normal[2] * current_tangential_1[1];
                            current_tangential_2[1] = current_normal[2] * current_tangential_1[0] - current_normal[0] * current_tangential_1[2];
                            current_tangential_2[2] = current_normal[0] * current_tangential_1[1] - current_normal[1] * current_tangential_1[0];
                        }
                    }

                    double value1 = -values[k] * current_normal[0] * contactsList->getMortarScale(interface_j);
                    double value2 = -values[k] * current_normal[1] * contactsList->getMortarScale(interface_j);
                    double value3 = -values[k] * current_normal[2] * contactsList->getMortarScale(interface_j);

                    B->insertGlobalValues(row,
                                          Teuchos::tuple<global_ordinal_type>(col1),
                                          Teuchos::tuple<double>(value1));
                    B->insertGlobalValues(row,
                                          Teuchos::tuple<global_ordinal_type>(col2),
                                          Teuchos::tuple<double>(value2));
                    B->insertGlobalValues(row,
                                          Teuchos::tuple<global_ordinal_type>(col3),
                                          Teuchos::tuple<double>(value3));

                    if (contactsList->isSticking(interface_j))
                    {
                        value1 = values[k] * current_tangential_1[0] * contactsList->getMortarScale(interface_j);
                        value2 = values[k] * current_tangential_1[1] * contactsList->getMortarScale(interface_j);
                        value3 = values[k] * current_tangential_1[2] * contactsList->getMortarScale(interface_j);

                        B->insertGlobalValues(row + 1,
                                              Teuchos::tuple<global_ordinal_type>(col1),
                                              Teuchos::tuple<double>(value1));
                        B->insertGlobalValues(row + 1,
                                              Teuchos::tuple<global_ordinal_type>(col2),
                                              Teuchos::tuple<double>(value2));
                        B->insertGlobalValues(row + 1,
                                              Teuchos::tuple<global_ordinal_type>(col3),
                                              Teuchos::tuple<double>(value3));

                        value1 = values[k] * current_tangential_2[0] * contactsList->getMortarScale(interface_j);
                        value2 = values[k] * current_tangential_2[1] * contactsList->getMortarScale(interface_j);
                        value3 = values[k] * current_tangential_2[2] * contactsList->getMortarScale(interface_j);

                        B->insertGlobalValues(row + 2,
                                              Teuchos::tuple<global_ordinal_type>(col1),
                                              Teuchos::tuple<double>(value1));
                        B->insertGlobalValues(row + 2,
                                              Teuchos::tuple<global_ordinal_type>(col2),
                                              Teuchos::tuple<double>(value2));
                        B->insertGlobalValues(row + 2,
                                              Teuchos::tuple<global_ordinal_type>(col3),
                                              Teuchos::tuple<double>(value3));
                    }
                    if (numDPN == 4)
                    {
                        global_ordinal_type row2 = row + numLMPN - 1;
                        global_ordinal_type col4 = Indices_int[k] * numDPN + 3;
                        B->insertGlobalValues(row2,
                                              Teuchos::tuple<global_ordinal_type>(col4),
                                              Teuchos::tuple<double>(values[k]));
                    }
                }

                delete values;
                delete Indices_int;
            }
        }

        delete values;
        delete Indices_int;
    }

    B->fillComplete(map->mapDofs, map->mapLagrangeDofs);

    return B;
}
