#include "wMedium.h"
#include "wProblem.h"
#include "wTag.h"
using namespace katropton;
using namespace tbox;

/**
 * @brief Medium constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> no: the ID of the body of the msh,
 *  <li> _medium_name: the name of new medium (does not have to be a nametag of the msh),
 *  <li> _E: a vector of values of the Young modulus \f$E\f$ (one per sample),
 *  <li> _nu: a vector of values of the Poisson coefficients \f$\nu\f$ (one per sample),
 *  <li> _k: the heat conductivity \f$k\f$ equal to zero by default,
 *  <li> _beta: the coefficient of thermal expansion \f$\beta\f$ equal to zero by default.
 * </ul>
 */
Medium::Medium(Problem &pbl,
               int no,
               std::string _medium_name,
               std::vector<double> _E,
               std::vector<double> _nu,
               std::vector<double> _k,
               std::vector<double> _beta) : Group(pbl.msh, no), medium_name(_medium_name)
{
    E = _E[0];
    nu = _nu[0];
    k = _k[0];
    beta = _beta[0];
    E_vector = _E;
    nu_vector = _nu;
    k_vector = _k;
    beta_vector = _beta;
    pbl.media.push_back(this);
}

/**
 * @brief Medium constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the body of the msh,
 *  <li> _medium_name: the name of new medium (does not have to be a nametag of the msh),
 *  <li> _E: a vector of values of the Young modulus \f$E\f$ (one per sample),
 *  <li> _nu: a vector of values of the Poisson coefficients \f$\nu\f$ (one per sample),
 *  <li> _k: the heat conductivity \f$k\f$ equal to zero by default,
 *  <li> _beta: the coefficient of thermal expansion \f$\beta\f$ equal to zero by default.
 * </ul>
 */
Medium::Medium(Problem &pbl,
               std::string const &name,
               std::string _medium_name,
               std::vector<double> _E,
               std::vector<double> _nu,
               std::vector<double> _k,
               std::vector<double> _beta) : Group(pbl.msh, name), medium_name(_medium_name)
{
    E = _E[0];
    nu = _nu[0];
    k = _k[0];
    beta = _beta[0];
    E_vector = _E;
    nu_vector = _nu;
    k_vector = _k;
    beta_vector = _beta;
    pbl.media.push_back(this);
}

/**
 * @brief Medium constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> no: the ID of the body of the msh,
 *  <li> _medium_name: the name of new medium (does not have to be a nametag of the msh),
 *  <li> _E: a vector of values of the Young modulus \f$E\f$ (one per sample),
 *  <li> _nu: a vector of values of the Poisson coefficients \f$\nu\f$ (one per sample),
 *  <li> _k: a vector of values of the heat conductivity \f$k\f$ (one per sample),
 *  <li> _beta: a vector of values of the coefficient of thermal expansion \f$\beta\f$ (one per sample).
 * </ul>
 */
Medium::Medium(Problem &pbl,
               int no,
               std::string _medium_name,
               std::vector<double> _E,
               std::vector<double> _nu,
               double _k,
               double _beta) : Group(pbl.msh, no), medium_name(_medium_name)
{
    E = _E[0];
    nu = _nu[0];
    k = _k;
    beta = _beta;
    E_vector = _E;
    nu_vector = _nu;
    std::vector<double> vect_k(E_vector.size(), _k);
    k_vector = vect_k;
    std::vector<double> vect_beta(E_vector.size(), _beta);
    beta_vector = vect_beta;
    pbl.media.push_back(this);
}

/**
 * @brief Medium constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the body of the msh,
 *  <li> _medium_name: the name of new medium (does not have to be a nametag of the msh),
 *  <li> _E: a vector of values of the Young modulus \f$E\f$ (one per sample),
 *  <li> _nu: a vector of values of the Poisson coefficients \f$\nu\f$ (one per sample),
 *  <li> _k: a vector of values of the heat conductivity \f$k\f$ (one per sample),
 *  <li> _beta: a vector of values of the coefficient of thermal expansion \f$\beta\f$ (one per sample).
 * </ul>
 */
Medium::Medium(Problem &pbl,
               std::string const &name,
               std::string _medium_name,
               std::vector<double> _E,
               std::vector<double> _nu,
               double _k,
               double _beta) : Group(pbl.msh, name), medium_name(_medium_name)
{
    E = _E[0];
    nu = _nu[0];
    k = _k;
    beta = _beta;
    E_vector = _E;
    nu_vector = _nu;
    std::vector<double> vect_k(E_vector.size(), _k);
    k_vector = vect_k;
    std::vector<double> vect_beta(E_vector.size(), _beta);
    beta_vector = vect_beta;
    pbl.media.push_back(this);
}

/**
 * @brief Display information.
 */
void Medium::write(std::ostream &out) const
{
    out << "Medium " << medium_name << " on " << *tag << " with E=" << this->E
        << ", nu=" << this->nu << ", k=" << this->k << " and beta=" << this->beta;
}
