/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WELEMENT_H
#define WELEMENT_H

#include "tbox.h"
#include "wPt.h"
#include "wObject.h"
#include <vector>
#include "gmm_gmm.h"

namespace tbox
{

/**
 * @brief element types (from gmsh)
 * @authors Romain Boman
 */
enum class ELTYPE
{
    UNKNOWN = 0,
    LINE2 = 1,
    TRI3 = 2,
    QUAD4 = 3,
    TETRA4 = 4,
    HEX8 = 5,
    POINT1 = 15
};
#ifndef SWIG
TBOX_API std::ostream &operator<<(std::ostream &out, ELTYPE const &obj);
#endif

enum class MATTYPE
{
    MAT_K,
    MAT_M,
    MAT_S
};
#ifndef SWIG
TBOX_API std::ostream &operator<<(std::ostream &out, MATTYPE const &obj);
#endif

/**
 * @brief Base finite element class
 * @authors Romain Boman, Kim Liegeois, Adrien Crovato
 */
class TBOX_API Element : public fwk::wObject
{
public:
    int no;                    ///< label
    Tag *ptag;                 ///< physical tag   (group)
    Tag *etag;                 ///< elementary tag (geometry)
    std::vector<Tag *> parts;  ///< partitions
    std::vector<Node *> nodes; ///< nodes

#ifndef SWIG
    Element(int n, Tag *_ptag, Tag *_etag, std::vector<Tag *> &_parts, std::vector<Node *> &nods);
    virtual ~Element() {}
#endif
    virtual ELTYPE type() const
    {
        return ELTYPE::UNKNOWN;
    }

    // test functions
    void buildK();
    void buildM();
    void buildS();
    void builds();

#ifndef SWIG
    virtual void gp2node(gmm::dense_matrix<double> &values_at_gp, gmm::dense_matrix<double> &values_at_node);
    virtual double buildJ(size_t k, gmm::dense_matrix<double> &J) const;
    virtual double buildJ(size_t k) const;
    //--------------------
    virtual void write(std::ostream &out) const override;
    virtual void buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct);
    virtual void buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct2 const &fct, bool fake);
    virtual void buildDK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct);
    //---------------------
    virtual void buildK_linElastic(gmm::dense_matrix<double> &K, gmm::dense_matrix<double> &H);
    virtual void buildK_mechanic(gmm::dense_matrix<double> &K_e, gmm::dense_matrix<double> &H);
    virtual void buildS_thermomechanic(gmm::dense_matrix<double> &S_e, gmm::dense_matrix<double> &D);
    virtual void buildL_thermic(gmm::dense_matrix<double> &L_e, gmm::dense_matrix<double> &K_conductivity);
    virtual void strain_stress(gmm::dense_matrix<double> &Strain, gmm::dense_matrix<double> &Stress, gmm::dense_matrix<double> &H, std::vector<double> &u);
    virtual void iso_coordinates(Pt x, double eta, double dzeta);
    virtual Pt normal();
    //---------------------
    virtual void buildM(gmm::dense_matrix<double> &M);
    virtual void buildS(gmm::dense_matrix<double> &S);
    virtual void builds(std::vector<double> &s, std::vector<double> const &u, Fct0 const &f);
    virtual void builds(std::vector<double> &s, std::vector<double> const &u, Fct1 const &f);
    virtual void buildDs(std::vector<double> &s, std::vector<double> const &u, Fct0 const &f);
    virtual void build(MATTYPE type, gmm::dense_matrix<double> &A);
    virtual void computeGrad(std::vector<double> const &u, std::vector<double> &grad, size_t k);
    virtual double computeGrad2(std::vector<double> const &u, size_t k);
    virtual void computeqV(std::vector<double> const &u, std::vector<double> &qV, Fct0 const &fct);
    virtual void computeqV(std::vector<double> const &u, std::vector<double> &qV, Fct2 const &fct);
    virtual void computeqint(std::vector<double> const &u, std::vector<double> &qV, Fct2 const &fct);
    virtual double computeV(std::vector<double> const &u, Fct0 const &fct);
    virtual void computeV(std::vector<double> const &u, gmm::dense_matrix<double> &out, Fct2 const &fct);
    virtual double computeV();

    virtual Cache &getVCache() = 0;
    virtual Mem &getVMem();
#endif
};

} // namespace tbox

#endif //WELEMENT_H
