/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wMemTetra4.h"
#include "wNode.h"
#include "wLazy.h"
#include "wFct0.h"
#include "gmm_gmm.h"

using namespace tbox;

Tetra4::Tetra4(int n, Tag *_ptag, Tag *_etag,
               std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{   
    pmem = new MemTetra4(*this, static_cast<int>(getCache().gauss.p.size()));
}
Tetra4::~Tetra4()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Tetra4 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem & Tetra4::getVMem()
{
    return getMem();
}
MemTetra4 & Tetra4::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Tetra4 
 * @authors Romain Boman, Adrien Crovato
 */
CacheTetra4 & Tetra4::getCache()
{
    static Lazy<CacheTetra4> cache(2);
    return cache.get();
}
Cache & Tetra4::getVCache()
{
    return getCache();
}

/**
 * @brief Compute Jacobian Matrix at Gauss point k
 * @authors Romain Boman
 */
double Tetra4::buildJ(size_t k, gmm::dense_matrix<double> &J) const
{
    CacheTetra4 &cache = getCache();

    gmm::dense_matrix<double> Jn(3, 3);
    gmm::clear(J);

    int i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];
        gmm::mult(gmm::col_vector(dff), gmm::row_vector((*it)->pos.x), Jn);
        gmm::add(Jn, J);
    }
    double detJ = gmm::lu_det(J);
    return detJ;
}

/**
 * @brief Compute mass matrix
 * @authors Romain Boman
 */
void Tetra4::buildM(gmm::dense_matrix<double> &M)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    gmm::clear(M);

    gmm::dense_matrix<double> Mk(nodes.size(), nodes.size());

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        //std::cout << "gp: " << k << " at " << gauss.p[k] << " with weight = " << gauss.w[k] << "\n";

        //Pt ksi(gauss.p[k]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(k);

        std::vector<double> &ff = cache.ff[k];
        gmm::mult(gmm::col_vector(ff), gmm::row_vector(ff), Mk); // Mk = {ff}^T [ff]    // idem pour tous les elements/pg "k" => cache?
        gmm::add(gmm::scaled(Mk, gauss.w[k] * detJ), M);         // M = M + Mk.wG.detJ
    }
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(4,4) = int fct*dN_j*dN_i dV
 * @authors Romain Boman, Adrien Crovato
 */
void Tetra4::buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();
    gmm::clear(K);
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluate function
        double fk = fct.eval(*this, u, k);
        // compute inverse Jacobian and determinant
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        // compute stiffness matrix for each GP
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));
                K(i, j) += gmm::vect_sp(factor1, factor2) * gauss.w[k] * detJ * fk;
            }
        }
    }
}

/**
 * @brief Compute (first part of) derivative stiffness matrix for scalar field (Newton method)
 * 
 * K_ij(4,4) = int (df du*dN_j du*dN_i) dV
 * @authors Adrien Crovato
 */
void Tetra4::buildDK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();
    gmm::clear(K);
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);

    // Gauss integration
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // "derivative" of f
        double dfk = fct.evalD(*this, u, k);

        // Jacobian
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        // gradient
        std::vector<double> gradu(3);
        computeGrad(u, gradu, k);

        // computation of K
        for (size_t i = 0; i < nodes.size(); ++i) {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            for (size_t j = 0; j < nodes.size(); ++j) {
                std::vector<double> &dffj = cache.dff[k][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));
                K(i, j) += gmm::vect_sp(gradu, factor1) * gmm::vect_sp(gradu, factor2) * gauss.w[k] * detJ * dfk;
            }
        }
    }
}

/**
 * @brief Compute derivative of flux vector for a scalar field (Newton method)
 * 
 * s_i(4) = int f*du*dN_i dV
 * @authors Adrien Crovato
 */
void Tetra4::buildDs(std::vector<double> &s, std::vector<double> const &u, Fct0 const &f)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();
    std::vector<double> factor1(3);
    gmm::clear(s);

    for (size_t k = 0; k < gauss.p.size(); ++k) {
        std::vector<double> gradu(3);
        this->computeGrad(u, gradu, k);
        double fk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        // s(i) = f * gradu * gradN(i)
        for (size_t i = 0; i < nodes.size(); ++i) {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            s[i] += gmm::vect_sp(gradu, factor1) * gauss.w[k] * detJ * fk;
        }
    }
}

/**
 * @brief Compute linear elasticity stiffness matrix
 * @authors Adrien Crovato
 * @todo this is a clean duplicate of buildK_mechnical, should be merged
 */
void Tetra4::buildK_linElastic(gmm::dense_matrix<double> &K, gmm::dense_matrix<double> &H)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();
    gmm::clear(K);
    std::vector<double> factori(3);

    for (size_t k = 0; k < gauss.p.size(); ++k) {
        // Determinant and inverse of Jacobian
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &invJ = mem.getJinv(k);

        // Fill B matrix
        gmm::dense_matrix<double> B(6, 3*nodes.size());
        for (size_t i = 0; i < nodes.size(); ++i) {
            // Compute [Ni,x Ni,y, Ni_z]
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(invJ, gmm::col_vector(dffi), gmm::col_vector(factori));
            B(0, 3*i) = factori[0];
            B(1, 3*i+1) = factori[1];
            B(2, 3*i+2) = factori[2];
            B(3, 3*i+1) = factori[2];
            B(3, 3*i+2) = factori[1];
            B(4, 3*i) = factori[2];
            B(4, 3*i+2) = factori[0];
            B(5, 3*i) = factori[1];
            B(5, 3*i+1) = factori[0];
        }

        // Compute stiffness matrix
        gmm::dense_matrix<double> HB(6, 3*nodes.size());
        gmm::dense_matrix<double> Kn(3*nodes.size(), 3*nodes.size());
        gmm::mult(H,B, HB);
        gmm::mult(gmm::transposed(B),HB, Kn);
        gmm::scale(Kn, detJ*gauss.w[k]);
        gmm::add(Kn, K);
    }
}

/* -------------------------------------------------------------------------- */
/**
 * @brief Compute mechanical stiffness matrix
 * @authors Kim Liegeois
 */
void Tetra4::buildK_mechanic(gmm::dense_matrix<double> &K_e, gmm::dense_matrix<double> &H)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    gmm::clear(K_e);
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);
    std::vector<double> factor3(3);

    gmm::dense_matrix<double> TMP1(3, 3);

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        //Pt ksi(gauss.p[a]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));

                gmm::clear(TMP1);
                for (size_t k = 0; k < 3; ++k)
                    for (size_t m = 0; m < 3; ++m)
                    {
                        for (size_t l = 0; l < 3; ++l)
                            for (size_t n = 0; n < 3; ++n)
                                TMP1(l, n) = H(k * 3 + l, m * 3 + n);

                        gmm::mult(TMP1, gmm::col_vector(factor2), gmm::col_vector(factor3));
                        K_e(i * 3 + k, j * 3 + m) += gmm::vect_sp(factor1, factor3) * gauss.w[a] * detJ;
                    }
            }
        }
    }

    for (size_t i = 0; i < 3 * nodes.size(); ++i)
        for (size_t j = i; j < 3 * nodes.size(); ++j)
        {
            if (fabs(K_e(i, j)) < epsilon)
                K_e(i, j) = 0;
            K_e(j, i) = K_e(i, j);
        }
}

/**
 * @brief Compute thermomechanical mass matrix
 * @authors Kim Liegeois
 */
void Tetra4::buildS_thermomechanic(gmm::dense_matrix<double> &S_e, gmm::dense_matrix<double> &D)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    gmm::clear(S_e); // dimension: 12*4
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        //Pt ksi(gauss.p[a]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        std::vector<double> &ff = cache.ff[a];

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            gmm::mult(D, gmm::col_vector(factor1), gmm::col_vector(factor2));
            for (size_t j = 0; j < nodes.size(); ++j)
                for (size_t k = 0; k < 3; ++k)
                    S_e(i * 3 + k, j) += factor2[k] * ff[j] * gauss.w[a] * detJ;
        }
    }

    for (size_t i = 0; i < 3 * nodes.size(); ++i)
        for (size_t j = 0; j < nodes.size(); ++j)
            if (fabs(S_e(i, j)) < epsilon)
                S_e(i, j) = 0;
}

/**
 * @brief Compute thermic matrix
 * @authors Kim Liegeois
 */
void Tetra4::buildL_thermic(gmm::dense_matrix<double> &L_e, gmm::dense_matrix<double> &K_conductivity)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    gmm::clear(L_e); // dimension: 4*4
    std::vector<double> factor1(3);
    std::vector<double> factor2(3);
    std::vector<double> factor3(3);

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        //Pt ksi(gauss.p[a]);

        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(a);
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        // calcul de K (au point de Gauss "ksi")

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            gmm::mult(K_conductivity, gmm::col_vector(factor1), gmm::col_vector(factor2));
            for (size_t j = i; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[a][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor3));
                L_e(i, j) += gmm::vect_sp(factor2, factor3) * gauss.w[a] * detJ;
            }
        }
    }

    for (size_t i = 0; i < nodes.size(); ++i)
        for (size_t j = i; j < nodes.size(); ++j)
        {
            if (fabs(L_e(i, j)) < epsilon)
                L_e(i, j) = 0;
            L_e(j, i) = L_e(i, j);
        }
}

/**
 * @brief Compute stress and strain
 * @authors Kim Liegeois
 */
void Tetra4::strain_stress(gmm::dense_matrix<double> &Strain, gmm::dense_matrix<double> &Stress, gmm::dense_matrix<double> &H, std::vector<double> &u)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    gmm::clear(Strain);
    gmm::clear(Stress);

    gmm::dense_matrix<double> Tmp(3, 3);
    gmm::clear(Tmp);

    gmm::dense_matrix<double> du_d_Phi(3, 3);
    gmm::dense_matrix<double> u_e(3, nodes.size());

    for (size_t i = 0; i < nodes.size(); ++i)
    {
        u_e(0, i) = u[3 * (nodes[i]->row)];
        u_e(1, i) = u[3 * (nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (nodes[i]->row) + 2];
    }

    // Machine epsilon

    double epsilon = 2e-16;

    for (size_t a = 0; a < gauss.p.size(); ++a)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        gmm::dense_matrix<double> const &J = mem.getJinv(a); // J contient maintenant Jinv!

        gmm::clear(du_d_Phi);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[a][i];
            for (auto k = 0; k < 3; ++k)
                for (auto j = 0; j < 3; ++j)
                    du_d_Phi(j, k) += dffi[j] * u_e(k, i);
        }

        for (size_t k = 0; k < 3; ++k)
            for (size_t m = 0; m < 3; ++m)
                for (size_t i = 0; i < 3; ++i)
                    Tmp(k, m) += J(k, i) * du_d_Phi(i, m) / (nodes.size());
    }
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            Strain(i, j) = (Tmp(i, j) + Tmp(j, i)) / 2;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            if (fabs(Strain(i, j)) < epsilon)
                Strain(i, j) = 0;

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);

    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            if (fabs(Stress(i, j)) < epsilon)
                Stress(i, j) = 0;
}

/* -------------------------------------------------------------------------- */

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 * @authors Romain Boman, Adrien Crovato
 */
void Tetra4::computeGrad(std::vector<double> const &u, std::vector<double> &grad, size_t k)
{
    CacheTetra4 &cache = getCache();
    MemTetra4 &mem = getMem();

    // local matrix of sf derivatives and solution vector
    gmm::dense_matrix<double> gradN(3, nodes.size());
    std::vector<double> ue(nodes.size());
    for (size_t i = 0; i < nodes.size(); ++i) {
        for (int j = 0; j < 3; ++j)
            gradN(j, i) = cache.dff[k][i][j];
        ue[i] = u[nodes[i]->row];
    }

    // gradient in element axis: d_jN_i * ue_i
    std::vector<double> grad_ue(3);
    gmm::mult(gradN, gmm::col_vector(ue), gmm::col_vector(grad_ue));

    // gradient in global axis: inv(J)_ij * grad_ue_j
    gmm::dense_matrix<double> const &Jinv = mem.getJinv(k);
    gmm::mult(Jinv, gmm::col_vector(grad_ue), gmm::col_vector(grad));
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 * @authors Adrien Crovato
 */
double Tetra4::computeGrad2(std::vector<double> const &u, size_t k)
{
    double grad2 = 0.;
    std::vector<double> grad(3);
    computeGrad(u, grad, k);
    for (size_t i = 0; i < grad.size(); ++i)
        grad2 += grad[i]*grad[i];

    return grad2;
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Adrien Crovato
 */
double Tetra4::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTetra4 &cache = getCache();
    GaussTetra4 &gauss = cache.gauss;
    MemTetra4 &mem = getMem();

    double V = 0.0;
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

/**
 * @brief Return the element volume
 * @authors Adrien Crovato
 */
double Tetra4::computeV()
{
    MemTetra4 &mem = getMem();
    return mem.getVol();
}
