/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTri3.h"
#include "wSfTri3.h"
#include "wGaussTri3.h"
#include "wCacheTri3.h"
#include "wMemTri3.h"
#include "wNode.h"
#include "wMath.h"
#include "wLazy.h"
#include "wFct0.h"
#include "wFct1.h"
#include "wFct2.h"
#include "gmm_gmm.h"

using namespace tbox;

Tri3::Tri3(int n, Tag *_ptag, Tag *_etag,
           std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemTri3(*this, static_cast<int>(getCache().gauss.p.size()));
}
Tri3::~Tri3()
{
    if (pmem)
        delete pmem;
}

/*
 * @brief Private memory handling Jacobian for Tri3 Element 
 * @authors Romain Boman, Adrien Crovato
 */
Mem &Tri3::getVMem()
{
    return getMem();
}
MemTri3 &Tri3::getMem()
{
    return *pmem;
}

/*
 * @brief Private cache handling shape functions and Gauss points for Tri3 
 * @authors Romain Boman, Adrien Crovato
 */
CacheTri3 &Tri3::getCache()
{
    static Lazy<CacheTri3> cache(2);
    return cache.get();
}
Cache &Tri3::getVCache()
{
    return getCache();
}

Pt Tri3::normal()
{
    Pt x1 = nodes[0]->pos;
    Pt x2 = nodes[1]->pos;
    Pt x3 = nodes[2]->pos;

    return ((x2 - x1) ^ (x3 - x1)).normalized();
}

/**
 * @brief Compute Jacobian determinant at Gauss point k
 * @authors Romain Boman
 * From Quad4, to check
 */
double Tri3::buildJ(size_t k) const
{
    CacheTri3 &cache = getCache();

    std::vector<double> tg1(3), tg2(3);
    gmm::clear(tg1);
    gmm::clear(tg2);

    int i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        std::vector<double> &dff = cache.dff[k][i];

        gmm::add(gmm::scaled((*it)->pos.x, dff[0]), tg1);
        gmm::add(gmm::scaled((*it)->pos.x, dff[1]), tg2);
    }

    std::vector<double> n(3);
    cross(tg1, tg2, n);
    double detJ = gmm::vect_norm2(n);
    return detJ;
}
/**
 * @brief Compute Jacobian matrix at Gauss point k
 * @authors Romain Boman
 */
double Tri3::buildJ(size_t k, gmm::dense_matrix<double> &J) const
{
    //std::cout << "\tTri3::buildJ...\n";
    CacheTri3 &cache = getCache();

    gmm::dense_matrix<double> Jn(2, 2);
    gmm::clear(J);

    int i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
    {
        //std::cout << "\ti=" << i << std::endl;
        std::vector<double> &dff = cache.dff[k][i];
        gmm::mult(gmm::col_vector(dff), gmm::row_vector(gmm::sub_vector((*it)->pos.x, gmm::sub_interval(0, 2))), Jn); // x is 3D so we take the first 2 componenents (x and y)
        gmm::add(Jn, J);
    }
    double detJ = gmm::lu_det(J);
    return detJ;
}

/**
 * @brief Compute stiffness matrix for scalar field
 * 
 * K_ij(3,3) = int fct*dN_j*dN_i dV
 * @authors Romain Boman
 */
void Tri3::buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();
    gmm::clear(K);
    std::vector<double> factor1(2);
    std::vector<double> factor2(2);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        //std::cout << "\tk=" << k << "\n";

        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);

        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        // calcul de K (au point de Gauss k)

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];

                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));
                K(i, j) += gmm::vect_sp(factor1, factor2) * gauss.w[k] * detJ * fk;
            }
        }
    }
}

void Tri3::buildK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct2 const &fct, bool fake)
{
    //std::cout << "\tTri3::buildK...\n";

    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;

    if (fake) // fake run - fill MPI job list
    {
        for (size_t k = 0; k < gauss.p.size(); ++k)
        {
            gmm::dense_matrix<double> fk(2, 2);
            fct.eval(*this, u, k, fk, fake);
        }
    }
    else // true run - use MPI cached results
    {
        MemTri3 &mem = getMem();
        gmm::clear(K);
        std::vector<double> factor0(2);
        std::vector<double> factor1(2);
        std::vector<double> factor2(2);

        for (size_t k = 0; k < gauss.p.size(); ++k)
        {
            //std::cout << "\tk=" << k << "\n";

            // evaluation de la fct a intégrer
            gmm::dense_matrix<double> fk(2, 2);
            fct.eval(*this, u, k, fk, fake);

            double detJ = mem.getDetJ(k);
            gmm::dense_matrix<double> const &J = mem.getJinv(k);

            // calcul de K (au point de Gauss k)

            for (size_t i = 0; i < nodes.size(); ++i)
            {
                std::vector<double> &dffi = cache.dff[k][i];
                gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor0));
                gmm::mult(fk, gmm::col_vector(factor0), gmm::col_vector(factor1));

                for (size_t j = 0; j < nodes.size(); ++j)
                {
                    std::vector<double> &dffj = cache.dff[k][j];

                    gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));
                    K(i, j) += gmm::vect_sp(factor1, factor2) * gauss.w[k] * detJ;
                }
            }
        }
    }
}

/**
 * @brief Compute mechanical stiffness matrix
 * 
 * K_ij(6,6) = int eps_k H_ijkl eps_l dV = B(6,3)H(3,3)B(3,6)
 * @authors Adrien Crovato
 */
void Tri3::buildK_linElastic(gmm::dense_matrix<double> &K, gmm::dense_matrix<double> &H)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();
    gmm::clear(K);
    std::vector<double> factori(2);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // Determinant and inverse of Jacobian
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &invJ = mem.getJinv(k);

        // Fill B matrix
        gmm::dense_matrix<double> B(3, 2 * nodes.size());
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            // Compute [Ni,x Ni,y]
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(invJ, gmm::col_vector(dffi), gmm::col_vector(factori));
            B(0, 2 * i) = factori[0];
            B(1, 2 * i + 1) = factori[1];
            B(2, 2 * i + 1) = factori[0];
            B(2, 2 * i) = factori[1];
        }

        // Compute stiffness matrix
        gmm::dense_matrix<double> HB(3, 2 * nodes.size());
        gmm::dense_matrix<double> Kn(2 * nodes.size(), 2 * nodes.size());
        gmm::mult(H, B, HB);
        gmm::mult(gmm::transposed(B), HB, Kn);
        gmm::scale(Kn, detJ * gauss.w[k]);
        gmm::add(Kn, K);
    }
}

/**
 * @brief Compute (first part of) derivative stiffness matrix for scalar field (Newton method)
 * 
 * K_ij(3,3) = int (df du*dN_j du*dN_i) dV
 * @authors Adrien Crovato
 */
void Tri3::buildDK(gmm::dense_matrix<double> &K, std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();
    gmm::clear(K);
    std::vector<double> factor1(2);
    std::vector<double> factor2(2);

    // Gauss integration
    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // "derivative" of f
        double dfk = fct.evalD(*this, u, k);

        // Jacobian
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        // gradient
        std::vector<double> gradu(2);
        this->computeGrad(u, gradu, k);

        // computation of K
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            for (size_t j = 0; j < nodes.size(); ++j)
            {
                std::vector<double> &dffj = cache.dff[k][j];
                gmm::mult(J, gmm::col_vector(dffj), gmm::col_vector(factor2));
                K(i, j) += gmm::vect_sp(gradu, factor1) * gmm::vect_sp(gradu, factor2) * gauss.w[k] * detJ * dfk;
            }
        }
    }
}

/**
 * @brief Compute flux vector for a scalar field
 * 
 * s_i(3) = int f*N_i dV
 * @authors Romain Boman
 */
void Tri3::builds(std::vector<double> &s, std::vector<double> const &u, Fct0 const &f)
{
    //std::cout << "\tTri3::builds...\n";

    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(s);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        double sk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        gmm::add(gmm::scaled(cache.ff[k], sk * gauss.w[k] * detJ), s); // s = s + sk.wG.detJ
    }
}

/**
 * @brief Compute flux vector for scalar field
 * 
 * s_i(3) = int f_j*n_j * N_i dV 
 * @authors Adrien Crovato
 */
void Tri3::builds(std::vector<double> &s, std::vector<double> const &u, Fct1 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();
    gmm::clear(s);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        double sk = f.eval(*this, u, k) * normal();
        double detJ = mem.getDetJ(k);
        gmm::add(gmm::scaled(cache.ff[k], sk * gauss.w[k] * detJ), s); // s = s + sk.wG.detJ
    }
}

/**
 * @brief Compute derivative of flux vector for a scalar field (Newton method)
 * 
 * s_i(3) = int f*du*dN_i dV
 * @authors Adrien Crovato
 */
void Tri3::buildDs(std::vector<double> &s, std::vector<double> const &u, Fct0 const &f)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();
    std::vector<double> factor1(2);
    gmm::clear(s);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        std::vector<double> gradu(2, 0.);
        this->computeGrad(u, gradu, k);
        double fk = f.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        // s(i) = f * gradu * gradN(i)
        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));
            s[i] += gmm::vect_sp(gradu, factor1) * gauss.w[k] * detJ * fk;
        }
    }
}

/**
 * @brief calcule un s(3) = int_V (grad N_j).vec(x,y) dV
 */
/*void Tri3::builds2(std::vector<double> &s, std::vector<double> const &u, Fct0 const &f)
{
    //std::cout << "\tTri3::builds...\n";

    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(s);

    for(int k=0; k<gauss.p.size(); ++k)
    {

        // calcul de gradT en k 
        std::vector<double> gradu(2);
        computeGrad(u, gradu, k);

        // evaluation de la fct a intégrer
        std::vector<double> &qk = fct.eval(*this, cache.ff, u, k); // q_k = q(T,gradT)



        double detJ = mem.getDetJ(k);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        for(int i=0; i<nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            s[i] += gmm::vect_sp(factor1, qk) * gauss.w[k] * detJ;
        }
    }
}*/

/**
 * @brief Compute something
 * 
 * S_ij(3,3) = int N_i*N_j dV (from Quad4)
 * @authors Romain Boman
 */
void Tri3::buildS(gmm::dense_matrix<double> &S)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(S);

    gmm::dense_matrix<double> Sk(nodes.size(), nodes.size());

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // calcul de la matrice jacobienne (au point de Gauss "ksi")
        double detJ = mem.getDetJ(k);

        std::vector<double> &ff = cache.ff[k];
        gmm::mult(gmm::col_vector(ff), gmm::row_vector(ff), Sk); // Sk = {ff}^T [ff]    // idem pour tous les elements/pg "k" => cache?

        gmm::add(gmm::scaled(Sk, gauss.w[k] * detJ), S); // S = S + Sk.wG.detJ
    }
}

void Tri3::build(MATTYPE type, gmm::dense_matrix<double> &A, std::vector<double> const &u, Fct0 const &fct)
{
    switch (type)
    {
    case MATTYPE::MAT_S:
        buildS(A);
        break;
    case MATTYPE::MAT_K:
        buildK(A, u, fct);
        break;
    default:
        std::stringstream out;
        out << "Tri3 cannot build " << type;
        throw std::runtime_error(out.str());
    }
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 * @authors Romain Boman, Adrien Crovato
 */
void Tri3::computeGrad(std::vector<double> const &u, std::vector<double> &grad, size_t k)
{
    //std::cout << "Tri3::computeGrad\n";
    CacheTri3 &cache = getCache();
    MemTri3 &mem = getMem();

    // local matrix of sf derivatives and solution vector
    gmm::dense_matrix<double> gradN(2, nodes.size());
    std::vector<double> ue(nodes.size());
    for (size_t i = 0; i < nodes.size(); ++i)
    {
        for (int j = 0; j < 2; ++j)
            gradN(j, i) = cache.dff[k][i][j];
        ue[i] = u[nodes[i]->row];
    }

    // gradient in element axis: d_jN_i * ue_i
    std::vector<double> grad_ue(2);
    gmm::mult(gradN, gmm::col_vector(ue), gmm::col_vector(grad_ue));

    // gradient in global axis: inv(J)_ij * grad_ue_j
    gmm::dense_matrix<double> const &Jinv = mem.getJinv(k);
    gmm::mult(Jinv, gmm::col_vector(grad_ue), gmm::col_vector(grad));
}

/**
 * @brief Compute square of gradient of a scalar field at Gauss point k
 * @authors Adrien Crovato
 */
double Tri3::computeGrad2(std::vector<double> const &u, size_t k)
{
    double grad2 = 0.;
    std::vector<double> grad(2);
    computeGrad(u, grad, k);
    for (size_t i = 0; i < grad.size(); ++i)
        grad2 += grad[i] * grad[i];

    return grad2;
}

/**
 *  calculation of int_V -gradu.fct dV
 */

void Tri3::computeqV(std::vector<double> const &u, std::vector<double> &qV, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(qV);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation du gradient
        std::vector<double> gradk(2);
        computeGrad(u, gradk, k);

        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);

        gmm::add(gmm::scaled(gradk, -fk * gauss.w[k] * detJ), qV); // qV = qV + fk.wG.detJ
    }
}

/**
 *  calculation of int_V flux grad N_j dV
 */

void Tri3::computeqint(std::vector<double> const &u, std::vector<double> &qV, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(qV);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation du gradient
        std::vector<double> gradk(2);
        computeGrad(u, gradk, k);

        // calcul du flux au pg k =>flux
        gmm::dense_matrix<double> fk(2, 2);
        fct.eval(*this, u, k, fk, false);
        double detJ = mem.getDetJ(k);
        std::vector<double> flux(2);
        gmm::mult(fk, gmm::col_vector(gradk), gmm::col_vector(flux));

        // calcul du gradient des fcts de forme et produit scalaire
        std::vector<double> factor1(2);
        gmm::dense_matrix<double> const &J = mem.getJinv(k);

        for (size_t i = 0; i < nodes.size(); ++i)
        {
            std::vector<double> &dffi = cache.dff[k][i];
            gmm::mult(J, gmm::col_vector(dffi), gmm::col_vector(factor1));

            qV[i] += gmm::vect_sp(factor1, flux) * gauss.w[k] * detJ;
        }
    }
}

/**
 *  calculation of int_V -[fct2]gradu dV
 */

void Tri3::computeqV(std::vector<double> const &u, std::vector<double> &qV, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(qV);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation du gradient
        std::vector<double> gradk(2);
        computeGrad(u, gradk, k);

        // evaluation de la fct a intégrer
        gmm::dense_matrix<double> fk(2, 2);
        fct.eval(*this, u, k, fk, false);
        double detJ = mem.getDetJ(k);

        std::vector<double> factor(2);
        gmm::mult(fk, gmm::col_vector(gradk), gmm::col_vector(factor));

        gmm::add(gmm::scaled(factor, -gauss.w[k] * detJ), qV); // qV = qV + fk.wG.detJ
    }
}

/**
 * @brief Compute intergal of function
 * 
 * I = int fct dV
 * @authors Romain Boman
 */
double Tri3::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    double V = 0.0;

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation de la fct a intégrer
        double fk = fct.eval(*this, u, k);
        double detJ = mem.getDetJ(k);
        V += fk * gauss.w[k] * detJ;
    }
    return V;
}

void Tri3::computeV(std::vector<double> const &u, gmm::dense_matrix<double> &out, Fct2 const &fct)
{
    CacheTri3 &cache = getCache();
    GaussTri3 &gauss = cache.gauss;
    MemTri3 &mem = getMem();

    gmm::clear(out);

    for (size_t k = 0; k < gauss.p.size(); ++k)
    {
        // evaluation de la fct a intégrer
        gmm::dense_matrix<double> fk(2, 2);
        fct.eval(*this, u, k, fk, false);
        double detJ = mem.getDetJ(k);

        gmm::add(gmm::scaled(fk, gauss.w[k] * detJ), out); // [out] = [out] + [fk].wG.detJ
    }
    //std::cout << out << "\n";
    //exit(1);
}

void Tri3::iso_coordinates(Pt x, double eta, double dzeta)
{
    std::vector<double> iso(2);
    std::vector<double> b(2);
    gmm::dense_matrix<double> A(2, 2);

    double detA;

    //std::vector<int> index_1 = {1,1,2}; // pas supporté en VC 2012!
    //std::vector<int> index_2 = {2,3,3}; // pas supporté en VC 2012!
    std::vector<int> index_1(3);
    index_1[0] = 1;
    index_1[1] = 1;
    index_1[2] = 2;
    std::vector<int> index_2(3);
    index_2[0] = 2;
    index_2[1] = 3;
    index_2[2] = 3;

    auto it0 = nodes.begin();
    auto it1 = it0 + 1;
    auto it2 = it0 + 2;

    Pt P1 = ((*it1)->pos - (*it0)->pos);
    Pt P2 = ((*it2)->pos - (*it0)->pos);
    Pt P3 = (x - (*it0)->pos);

    for (size_t i = 0; i < 3; ++i)
    {
        A(0, 0) = P1.x[index_1[i]];
        A(1, 0) = P1.x[index_2[i]];

        A(0, 1) = P2.x[index_1[i]];
        A(1, 1) = P2.x[index_2[i]];

        detA = gmm::lu_det(A);

        if (detA == 0)
            continue;
        else
        {
            b[0] = P3.x[index_1[i]];
            b[1] = P3.x[index_2[i]];
            break;
        }
    }

    gmm::lu_solve(A, iso, b);

    if (iso[0] < 0 || iso[1] > 1 || iso[0] + iso[1] > 1)
        eta = -2;
    else
    {
        eta = iso[0];
        dzeta = iso[1];
    }
}

/**
 * @brief Return the element volume
 * @authors Romain Boman
 */
double Tri3::computeV()
{
    MemTri3 &mem = getMem();
    return mem.getVol();
}
