/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wVtkExport_KIM2CLEAN.h"
#include "wMshData.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include "wResults.h"

#include <vtkVersion.h>
#include <vtkSmartPointer.h>
#include <vtkTetra.h>
#include <vtkHexahedron.h>
#include <vtkCellArray.h>
#include <vtkXMLUnstructuredGridReader.h>

#include <vtkXMLUnstructuredGridWriter.h>
#include "vtkXMLPUnstructuredGridWriter.h"
#include <vtkUnstructuredGrid.h>
#include <vtkPointData.h>
#include <vtkDoubleArray.h>
#include "vtkCellArray.h"
#include "vtkCellData.h"
#include <vtkVertexGlyphFilter.h>

#ifdef VTK_HAS_MPI
#include <mpi.h>
#include "vtkMPIController.h"
#endif

using namespace tbox;
using namespace tboxVtk;

VtkExport_KIM2CLEAN::VtkExport_KIM2CLEAN(std::shared_ptr<MshData> _msh) : msh(_msh)
{
}

/**
 * @brief Save results defined at all the slave contact nodes ONLY.
 * Those datas are Lagrange multipliers and can only be scalar or vectors.
 */

void 
VtkExport_KIM2CLEAN::save_MPI(std::string const &fname, Results const &r, std::vector<int> contactElementID, std::map<int,int> lagrange_LO_to_nodes_GO, std::map<int,int> nodes_GO_to_lagrange_LO) const
{
#ifdef VTK_HAS_MPI
    vtkSmartPointer<vtkPoints> points =
        vtkSmartPointer<vtkPoints>::New();

    points->SetNumberOfPoints(lagrange_LO_to_nodes_GO.size());

    for(int ii=0; ii<lagrange_LO_to_nodes_GO.size(); ++ii)
        points->InsertPoint(ii,msh->nodes[lagrange_LO_to_nodes_GO[ii]]->pos.x[0],msh->nodes[lagrange_LO_to_nodes_GO[ii]]->pos.x[1],msh->nodes[lagrange_LO_to_nodes_GO[ii]]->pos.x[2]);


    int CellArraySize = 0;
    int numberOfCells = 0;
    int numberOfElems = contactElementID.size();
    bool isQuad = true;
    for(int ii=0; ii<numberOfElems; ++ii)
    {
        int elemID = contactElementID[ii];
        if(msh->elems[elemID]->type() == ELTYPE::QUAD4)
        {
            CellArraySize += ((msh->elems[elemID]->nodes.size()) + 1);
            ++numberOfCells;
        }
        else if(msh->elems[elemID]->type() == ELTYPE::TRI3)
        {
            CellArraySize += ((msh->elems[elemID]->nodes.size()) + 1);
            ++numberOfCells;
            isQuad = false;
        }        
    }

    vtkIdType *vtkCells = NULL;
    vtkCells = new vtkIdType[CellArraySize];

    int CellOffset = 0;
    for(int ii=0; ii<numberOfElems; ++ii)
    {
        int elemID = contactElementID[ii];
        if(msh->elems[elemID]->type() == ELTYPE::QUAD4)
        {
            vtkCells[CellOffset++] = msh->elems[elemID]->nodes.size();
            for(int jj=0; jj<msh->elems[elemID]->nodes.size(); ++jj)
                vtkCells[CellOffset++] = nodes_GO_to_lagrange_LO[msh->elems[elemID]->nodes[jj]->row];
        }
        else if(msh->elems[elemID]->type() == ELTYPE::TRI3)
        {
            vtkCells[CellOffset++] = msh->elems[elemID]->nodes.size();
            for(int jj=0; jj<msh->elems[elemID]->nodes.size(); ++jj)
                vtkCells[CellOffset++] = nodes_GO_to_lagrange_LO[msh->elems[elemID]->nodes[jj]->row];
        }        
    }
    vtkSmartPointer <vtkIdTypeArray> Conn = vtkSmartPointer<vtkIdTypeArray>::New();
    Conn->SetArray(vtkCells, CellArraySize,1);

    vtkSmartPointer<vtkCellArray> cellArray =
        vtkSmartPointer<vtkCellArray>::New();        
    cellArray->SetCells	(numberOfCells,Conn);

    vtkSmartPointer<vtkUnstructuredGrid> unstructuredGrid =
        vtkSmartPointer<vtkUnstructuredGrid>::New();
    unstructuredGrid->SetPoints(points);
    if (isQuad)
        unstructuredGrid->SetCells(VTK_QUAD, cellArray);
    else    
        unstructuredGrid->SetCells(VTK_TRIANGLE, cellArray);  

    for(auto &p : r.scalars_at_nodes)
    {
        vtkSmartPointer<vtkDoubleArray> scalar =
            vtkSmartPointer<vtkDoubleArray>::New();      

        scalar->SetNumberOfComponents(1);
        scalar->SetNumberOfTuples(lagrange_LO_to_nodes_GO.size());
        scalar->SetName(p.first.c_str());  

        for(int ii=0; ii<lagrange_LO_to_nodes_GO.size(); ++ii)
        {
            scalar->SetValue(ii, (*(p.second))[ii]);
        }

        unstructuredGrid->GetPointData()->AddArray(scalar);    
    }
    for(auto &p : r.vectors_at_nodes)
    {
        vtkSmartPointer<vtkDoubleArray> vector =
            vtkSmartPointer<vtkDoubleArray>::New();      

        vector->SetNumberOfComponents(3);
        vector->SetNumberOfTuples(lagrange_LO_to_nodes_GO.size());
        vector->SetName(p.first.c_str());  

        for(int ii=0; ii<lagrange_LO_to_nodes_GO.size(); ++ii)
        {
            Pt const &v = (*(p.second))[ii];
            vector->SetTuple3(ii, v.x[0], v.x[1], v.x[2]);
        }

        unstructuredGrid->GetPointData()->AddArray(vector);
    }

    vtkSmartPointer<vtkXMLUnstructuredGridWriter> writer =
        vtkSmartPointer<vtkXMLUnstructuredGridWriter>::New();
    writer->SetFileName(fname.c_str());
    writer->SetInputData(unstructuredGrid);
    writer->Write();  

    delete[] vtkCells;
#endif          
}

/**
 * @brief Save results defined at all nodes and 3D elements, those values can scalar, vector or tensors.
 */

void 
VtkExport_KIM2CLEAN::save_MPI(std::string const &fname, Results const &r, std::map<int,int> nodesWO_LO_to_GO, std::map<int,int> nodesWO_GO_to_LO , std::map<int,int> elements_GO_to_LO ) const
{
#ifdef VTK_HAS_MPI

    int n_procs = 0, rank = 0;
    MPI_Comm_size(MPI_COMM_WORLD, &n_procs);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    vtkMPIController* controller = vtkMPIController::New();
    controller->Initialize(0, 0, 1);
    controller->SetGlobalController(controller);

    vtkSmartPointer<vtkPoints> points =
        vtkSmartPointer<vtkPoints>::New();
  
    points->SetNumberOfPoints(nodesWO_LO_to_GO.size());

    for(int ii=0; ii<nodesWO_LO_to_GO.size(); ++ii)
        points->InsertPoint(ii,msh->nodes[nodesWO_LO_to_GO[ii]]->pos.x[0],msh->nodes[nodesWO_LO_to_GO[ii]]->pos.x[1],msh->nodes[nodesWO_LO_to_GO[ii]]->pos.x[2]);

    int CellArraySize = 0;
    int numberOfCells = 0;
    int numberOfElems = msh->elems.size();
    bool isAHexahedron = true;
  
    std::map<int,int> elems_GO_to_3D;
    std::map<int,int> elems_3D_to_LO;
  
    for(int ii=0; ii<numberOfElems; ++ii)
    {
        if(msh->elems[ii]->type() == ELTYPE::HEX8)
        {
            CellArraySize += ((msh->elems[ii]->nodes.size()) + 1);
          
            elems_GO_to_3D[msh->elems[ii]->no-1] = numberOfCells;
            elems_3D_to_LO[numberOfCells] = ii;
          
            ++numberOfCells;
        }
        else if(msh->elems[ii]->type() == ELTYPE::TETRA4)
        {
            CellArraySize += ((msh->elems[ii]->nodes.size()) + 1);
          
            elems_GO_to_3D[msh->elems[ii]->no-1] = numberOfCells;
            elems_3D_to_LO[numberOfCells] = ii;
          
            ++numberOfCells;
          
            isAHexahedron = false;
        }        
    }

    vtkIdType *vtkCells = NULL;
    vtkCells = new vtkIdType[CellArraySize];

    int CellOffset = 0;
    for(int ii=0; ii<numberOfElems; ++ii)
    {
        if(msh->elems[ii]->type() == ELTYPE::HEX8)
        {
            vtkCells[CellOffset++] = msh->elems[ii]->nodes.size();
            for(int jj=0; jj<msh->elems[ii]->nodes.size(); ++jj)
                vtkCells[CellOffset++] = nodesWO_GO_to_LO[msh->elems[ii]->nodes[jj]->row];
        }
        else if(msh->elems[ii]->type() == ELTYPE::TETRA4)
        {
            vtkCells[CellOffset++] = msh->elems[ii]->nodes.size();
            for(int jj=0; jj<msh->elems[ii]->nodes.size(); ++jj)
                vtkCells[CellOffset++] = nodesWO_GO_to_LO[msh->elems[ii]->nodes[jj]->row];
        }        
    }
    vtkSmartPointer <vtkIdTypeArray> Conn = vtkSmartPointer<vtkIdTypeArray>::New();
    Conn->SetArray(vtkCells, CellArraySize,1);

    vtkSmartPointer<vtkCellArray> cellArray =
        vtkSmartPointer<vtkCellArray>::New();    

    cellArray->SetCells	(numberOfCells,Conn);

    vtkSmartPointer<vtkUnstructuredGrid> unstructuredGrid =
        vtkSmartPointer<vtkUnstructuredGrid>::New();
    unstructuredGrid->SetPoints(points);
    if (isAHexahedron)
        unstructuredGrid->SetCells(VTK_HEXAHEDRON, cellArray);
    else    
        unstructuredGrid->SetCells(VTK_TETRA, cellArray);

    for(auto &p : r.scalars_at_nodes)
    {
        vtkSmartPointer<vtkDoubleArray> scalar =
            vtkSmartPointer<vtkDoubleArray>::New();      

        scalar->SetNumberOfComponents(1);
        scalar->SetNumberOfTuples(nodesWO_LO_to_GO.size());
        scalar->SetName(p.first.c_str());

        for(int ii=0; ii<nodesWO_LO_to_GO.size(); ++ii)
            scalar->SetValue(ii, (*(p.second))[ii]);  

        unstructuredGrid->GetPointData()->AddArray(scalar);    
    }
    for(auto &p : r.vectors_at_nodes)
    {
        vtkSmartPointer<vtkDoubleArray> vector =
            vtkSmartPointer<vtkDoubleArray>::New();      

        vector->SetNumberOfComponents(3);
        vector->SetNumberOfTuples(nodesWO_LO_to_GO.size());
        vector->SetName(p.first.c_str());

        for(int ii=0; ii<nodesWO_LO_to_GO.size(); ++ii)
        {
            Pt const &v = (*(p.second))[ii];
            vector->SetTuple3(ii, v.x[0], v.x[1], v.x[2]); 
        }

        unstructuredGrid->GetPointData()->AddArray(vector);
    }
    for(auto &p : r.tensors_at_nodes)
    {
        vtkSmartPointer<vtkDoubleArray> tensors =
            vtkSmartPointer<vtkDoubleArray>::New();

        tensors->SetNumberOfComponents(9);
        tensors->SetNumberOfTuples(nodesWO_LO_to_GO.size());
        tensors->SetName(p.first.c_str());

        for(int ii=0; ii<nodesWO_LO_to_GO.size(); ++ii)
        {
            gmm::dense_matrix<double> const &v = (*(p.second))[ii];
            tensors->SetTuple9(ii, v(0,0), v(0,1), v(0,2), v(1,0), v(1,1), v(1,2), v(2,0), v(2,1), v(2,2));
        }

        unstructuredGrid->GetPointData()->AddArray(tensors);
    }
  
  
    for(auto &p : r.scalars_at_elems)
    {
        vtkSmartPointer<vtkDoubleArray> scalar =
            vtkSmartPointer<vtkDoubleArray>::New();

        scalar->SetNumberOfComponents(1);
        scalar->SetNumberOfTuples(numberOfCells);
        scalar->SetName(p.first.c_str());

        for(int ii=0; ii<numberOfCells; ++ii)
        {
            scalar->SetValue(ii, (*(p.second))[elems_3D_to_LO[ii]]);
        }

        unstructuredGrid->GetCellData()->AddArray(scalar);
    }
    for(auto &p : r.vectors_at_elems)
    {
        vtkSmartPointer<vtkDoubleArray> vector =
            vtkSmartPointer<vtkDoubleArray>::New();

        vector->SetNumberOfComponents(3);
        vector->SetNumberOfTuples(numberOfCells);
        vector->SetName(p.first.c_str());

        for(int ii=0; ii<numberOfCells; ++ii)
        {
            Pt const &v = (*(p.second))[ii];
            vector->SetTuple3(ii, v.x[0], v.x[1], v.x[2]);
        }

        unstructuredGrid->GetCellData()->AddArray(vector);
    }
    for(auto &p : r.tensors_at_elems)
    {
        vtkSmartPointer<vtkDoubleArray> tensors =
            vtkSmartPointer<vtkDoubleArray>::New();

        tensors->SetNumberOfComponents(9);
        tensors->SetNumberOfTuples(numberOfCells);
        tensors->SetName(p.first.c_str());

        for(int ii=0; ii<numberOfCells; ++ii)
        {
            gmm::dense_matrix<double> const &v = (*(p.second))[elems_3D_to_LO[ii]];
            tensors->SetTuple9(ii, v(0,0), v(0,1), v(0,2), v(1,0), v(1,1), v(1,2), v(2,0), v(2,1), v(2,2));
        }

        unstructuredGrid->GetCellData()->AddArray(tensors);
    }
 
    vtkSmartPointer<vtkIntArray> elem_proc_id = vtkSmartPointer<vtkIntArray>::New();
    elem_proc_id->SetName("processor_id");
    elem_proc_id->SetNumberOfComponents(1);
    elem_proc_id->SetNumberOfTuples(numberOfCells);

    for(int ii=0; ii<numberOfCells; ++ii)
        elem_proc_id->SetValue(ii,rank); 

    unstructuredGrid->GetCellData()->AddArray(elem_proc_id);

    vtkSmartPointer<vtkIntArray> volume_id = vtkSmartPointer<vtkIntArray>::New();
    volume_id->SetName("volume_id");
    volume_id->SetNumberOfComponents(1);
    volume_id->SetNumberOfTuples(numberOfElems);

    for (auto& ptag : msh->ptags) {
        if (ptag.second->dim == 3){
            for(int ii=0; ii<ptag.second->elems.size(); ++ii){
                volume_id->SetValue(elems_GO_to_3D[ptag.second->elems[ii]->no-1],ptag.second->no);
            } 
        }
    }
    unstructuredGrid->GetCellData()->AddArray(volume_id);

    vtkSmartPointer<vtkXMLPUnstructuredGridWriter> writer =
        vtkSmartPointer<vtkXMLPUnstructuredGridWriter>::New();

    writer->SetGhostLevel(1);
    writer->SetNumberOfPieces(n_procs);
    writer->SetStartPiece(rank);
    writer->SetEndPiece(rank);

    writer->SetFileName(fname.c_str());
    writer->SetInputData(unstructuredGrid);
    writer->Write();  

    delete[] vtkCells;                      

    controller->Finalize(1);
#endif
}
