/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wForwardEuler.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTimer.h"
#include "wTag.h"
#include "wProblem.h"
#include "wDisplayHook.h"
#include "wSource.h"
#include "wResults.h"
#include "wMshExport.h"
#include "gmm_gmm.h"
using namespace tbox;
using namespace waves;

ForwardEuler::ForwardEuler(std::shared_ptr<waves::Problem> _pbl) : TimeIntegration(_pbl)
{
}

void ForwardEuler::start(std::shared_ptr<MshExport> mshWriter)
{
    std::cout << *this;
    chrono1.start();

    auto msh = pbl->msh;
    u1.resize(msh->nodes.size());
    v1.resize(msh->nodes.size());

    // assign a row for each node
    int i = 0;
    for (auto n : msh->nodes)
        n->row = i++;
    //std::for_each(msh->nodes.begin(), msh->nodes.end(), [&](Node *n) { n->row = i++; });

    // matrix assembly
    gmm::csr_matrix<double> K;
    std::vector<double> Md(msh->nodes.size());
    buildKM(K, Md, u1);

    gmm::csr_matrix<double> S;
    buildS(S);

    // setup results
    Results results;
    results.scalars_at_nodes["u"] = &u1;
    //results.scalars_at_nodes["v"] = &v1;

    // time integration

    std::cout << "\n\nTime integration...\n";
    std::cout << *this << '\n';

    std::vector<double> tmp1(msh->nodes.size()), tmp2(msh->nodes.size());
    double t = 0.0;
    int nt = 1;
    results.nt = nt;
    results.time = t;

    while (t < ttot && !stopit)
    {
        //if(verbose)    std::cout << "nt " << nt << " t=" << t << "\n";
        if (verbose)
            std::cout << "nt " << nt << " t=" << t << " stopit=" << stopit << "\n";

        //std::for_each(msh->srcs.begin(), msh->srcs.end(), [&](Source *s) { s->apply(t, u1); });
        for (auto s : pbl->srcs)
            s->apply(t, u1);

        t += dt;
        nt++;

        // equation #1
        gmm::add(u1, gmm::scaled(v1, dt), u1); // u1 = u0 + dt*v0

        for (auto s : pbl->srcs)
            s->apply(t, u1);

        // equation #2
        gmm::mult(K, u1, tmp1);
        gmm::mult(S, v1, tmp2);
        for (size_t i = 0; i < gmm::vect_size(Md); ++i)
            v1[i] = v1[i] - dt / Md[i] * (tmp1[i] + tmp2[i]);

        if (nt % savefreq == 0)
        {
            //std::cout << "nt " << nt << " t=" << t << "\n";
            results.time = t;
            if (wgmsh == Wgmsh::WG_ONE)
            {
                results.nt++;
                mshWriter->save("res", results);
            }
            else if (wgmsh == Wgmsh::WG_ALL)
            {
                results.nt++;
                char filename[256];
                sprintf(filename, "res_%06d", nt);
                mshWriter->save(filename, results);
            }
            if (dhook)
                dhook->display(nt, t, u1);
        }
        if (dhook)
            dhook->refresh();
    }

    chrono1.stop();
    std::cout << "[cpu] " << chrono1 << '\n';
}
